"""Test fitting and simulating recall."""

import numpy as np
import pytest
from psifr import fr

from cymr.fit import Recall
from cymr import fit
from cymr import parameters


class TestRecall(Recall):
    def prepare_sim(self, data, study_keys=None, recall_keys=None):
        data_study = data.loc[data['trial_type'] == 'study']
        data_recall = data.loc[data['trial_type'] == 'recall']
        merged = fr.merge_lists(data_study, data_recall)
        study = fr.split_lists(merged, 'study', ['input'])
        recalls = fr.split_lists(merged, 'recall', ['input'])
        return study, recalls

    def likelihood_subject(self, study, recalls, param, param_def=None, patterns=None):
        p = 2 - (param['x'] + 2) ** 2
        eps = 0.0001
        if p < eps:
            p = eps
        n = 1
        return np.log(p), n

    def generate_subject(
        self, study, recall, param, param_def=None, patterns=None, **kwargs
    ):
        recalls_list = [param['recalls']]
        return recalls_list

    def record_subject(self, study, recall, param, param_def=None, patterns=None):
        study_state = study['input']
        recall_state = recall['input']
        return study_state, recall_state


@pytest.fixture()
def data():
    data = fit.sample_data('sample1')
    return data


def test_likelihood_subject(data):
    data = data.copy()
    rec = TestRecall()
    subject_data = data.loc[data['subject'] == 1]
    study, recall = rec.prepare_sim(subject_data)
    param = {'x': -2}
    logl, n = rec.likelihood_subject(study, recall, param)
    np.testing.assert_allclose(logl, np.log(2))


def test_likelihood(data):
    data = data.copy()
    rec = TestRecall()
    param = {'x': -2}
    stats = rec.likelihood(data, param)
    np.testing.assert_allclose(stats['logl'].to_numpy(), [np.log(2), np.log(2)])


def test_fit_subject(data):
    data = data.copy()
    rec = TestRecall()
    param_def = parameters.Parameters()
    param_def.set_fixed(y=1)
    param_def.set_free(x=[-10, 10])
    subject_data = data.loc[data['subject'] == 1]
    param, logl, n, k = rec.fit_subject(subject_data, param_def)
    np.testing.assert_allclose(param['x'], -2, atol=0.00001)
    np.testing.assert_allclose(logl, np.log(2))
    assert k == 1


def test_fit_indiv(data):
    data = data.copy()
    rec = TestRecall()
    param_def = parameters.Parameters()
    param_def.set_fixed(y=1)
    param_def.set_free(x=[-10, 10])
    results = rec.fit_indiv(data, param_def)
    np.testing.assert_allclose(results['x'].to_numpy(), [-2, -2], atol=0.0001)
    np.testing.assert_allclose(results['logl'].to_numpy(), np.log([2, 2]), atol=0.0001)
    np.testing.assert_array_equal(results['k'].to_numpy(), [1, 1])


def test_generate_subject(data):
    data = data.copy()
    rec = TestRecall()
    study = data.loc[(data['trial_type'] == 'study') & (data['subject'] == 1)]
    param = {'recalls': ['hollow', 'pupil']}
    # our "model" recalls the positions indicated in the recalls parameter
    rec_list = rec.generate_subject(study, None, param)
    data_sim = fit.add_recalls(study, rec_list)
    expected = ['absence', 'hollow', 'pupil', 'hollow', 'pupil']
    assert data_sim['item'].to_list() == expected


def test_generate(data):
    data = data.copy()
    rec = TestRecall()
    subj_param = {
        1: {'recalls': ['hollow', 'pupil']},
        2: {'recalls': ['pillow', 'fountain', 'piano']},
    }
    sim = rec.generate(data, {}, subj_param)
    expected = [
        'absence',
        'hollow',
        'pupil',
        'hollow',
        'pupil',
        'fountain',
        'piano',
        'pillow',
        'pillow',
        'fountain',
        'piano',
    ]
    assert sim['item'].to_list() == expected


def test_record_subject(data):
    rec = TestRecall()
    study, recall = fit.prepare_lists(data.loc[data['subject'] == 1])
    study_state, recall_state = rec.record_subject(study, recall, {})
    np.testing.assert_array_equal(study_state[0], np.array([0, 1, 2]))
    np.testing.assert_array_equal(recall_state[0], np.array([1, 2]))


def test_record(data):
    rec = TestRecall()
    states = rec.record(data, {})
    np.testing.assert_array_equal(
        np.asarray(states), np.array([1, 2, 3, 2, 3, np.nan, 1, 2, 3, 3, 1, 3])
    )



"""Manage model parameter settings."""

from __future__ import annotations
import json
import copy

import numpy as np
from typing import Union, Any, Callable, Iterable, Optional
from numpy.typing import ArrayLike


def read_json(json_file: str) -> Parameters:
    """Read parameters from a JSON file."""
    with open(json_file, 'r') as f:
        par_dict = json.load(f)

    par = Parameters()
    if 'options' in par_dict:
        par.set_options(par_dict['options'])
    par.set_free(par_dict['free'])
    par.set_fixed(par_dict['fixed'])
    par.set_dependent(par_dict['dependent'])
    for trial_type, p in par_dict['dynamic'].items():
        for name, val in p.items():
            if 'scope' in par_dict['dynamic'] and name in par_dict['dynamic']['scope']:
                scope = par_dict['dynamic']['scope'][name]
            else:
                scope = 'item'
            par.set_dynamic(trial_type, scope, {name: val})
    return par


def set_dependent(
    param: dict[str, float], dependent: dict[str, str] = None
) -> dict[str, float]:
    """
    Set values of dependent parameters.

    Parameters
    ----------
    param : dict of (str: float)
        Parameter values.

    dependent : dict of (str: str)
        For each dependent parameter, an expression defining that
        parameter in terms of other parameter values.

    Returns
    -------
    updated : dict of (str: float)
        Updated parameter values.
    """
    updated = param.copy()
    if dependent is not None:
        for name, expression in dependent.items():
            updated[name] = eval(expression, np.__dict__, updated)
    return updated


def set_dynamic(
    param: dict[str, float],
    list_data: dict[str, list[ArrayLike]],
    dynamic: dict[str, str],
    scope: dict[str, str] = None,
) -> dict[str, Union[float, list[ArrayLike]]]:
    """
    Set dynamic parameters for one trial type.

    Parameters
    ----------
    param : dict of (str: float)
        Parameter values.

    list_data : dict of list of numpy.array
        Data in list format with named fields.

    dynamic : dict of (str: str)
        Dynamic parameter definitions for one trial type. Expressions
        will be evaluated with both param keys and data keys available
        as variables. If a key exists on both param and data, the param
        value takes precedence.
    
    scope : dict of (str: str)
        Scope of parameters. If "list", the unique value will be taken.

    Returns
    -------
    updated : dict
        Parameter values updated to include dynamic parameters.
        Dynamic parameters are lists of numpy arrays.
    """
    # flip from dict of list to list of dict
    data_keys = list(list_data.keys())
    n_list = len(list_data[data_keys[0]])
    data_list = [{key: val[i] for key, val in list_data.items()} for i in range(n_list)]

    # merge in parameters
    for i in range(len(data_list)):
        data_list[i].update(param)
    updated: dict[str, Any] = param.copy()

    # set each dynamic parameter
    for key, expression in dynamic.items():
        updated[key] = []
        for data in data_list:
            # restrict eval functions to the numpy namespace
            val = eval(expression, np.__dict__, data)
            if scope is not None and key in scope and scope[key] == 'list':
                uval = np.unique(val)
                if len(uval) > 1:
                    raise ValueError(
                        f'Expression for "{key}", "{expression}", evaluated to multiple values.'
                    )
                val = float(uval[0])
            updated[key].append(val)
    return updated


def sample_parameters(
    sampler: dict[str, Union[float, Callable[[], float], tuple[float, float]]]
) -> dict[str, float]:
    """Randomly sample parameters."""
    param = {}
    for name, value in sampler.items():
        if callable(value):
            param[name] = value()
        elif isinstance(value, tuple):
            param[name] = value[0] + np.random.rand() * value[1]
        else:
            param[name] = value
    return param


class Parameters(object):
    """
    Class to manage model parameters.

    Attributes
    ----------
    fixed : dict of (str: float)
        Values of fixed parameters.

    free : dict of (str: tuple)
        Bounds of each free parameter.

    dependent : dict of (str: str)
        Expressions to define dependent parameters based the other
        parameters.

    dynamic : dict of (str: dict of (str: str))
        First dict specifies trial_type for dynamic parameters,
        second dict keys are parameter names, and values are
        expressions specifying how to update the parameter.
    """

    def __init__(self) -> None:
        self.options: dict[str, Any] = {}
        self.fixed: dict[str, float] = {}
        self.free: dict[str, Iterable[float]] = {}
        self.dependent: dict[str, str] = {}
        self.dynamic: dict[str, dict[str, str]] = {}
        self._dynamic_names: set[str] = set()
        self._fields: list[str] = [
            'fixed',
            'free',
            'dependent',
            'dynamic',
        ]

    def __repr__(self) -> str:
        parts: dict[str, str] = {}
        for name in self._fields:
            obj = getattr(self, name)
            fields = [f'{key}: {value}' for key, value in obj.items()]
            parts[name] = '\n'.join(fields)
        s = '\n\n'.join([f'{name}:\n{f}' for name, f in parts.items()])
        return s

    def copy(self) -> Parameters:
        """Copy the parameters definition."""
        return copy.deepcopy(self)

    def to_json(self, json_file: str) -> None:
        """
        Write parameter definitions to a JSON file.

        Parameters
        ----------
        json_file : str
            Path to file to save json data.
        """
        data: dict[str, Any] = {
            'options': self.options,
            'fixed': self.fixed,
            'free': self.free,
            'dependent': self.dependent,
            'dynamic': self.dynamic,
        }
        with open(json_file, 'w') as f:
            json.dump(data, f, indent=4)

    def set_options(self, *args: dict[str, Any], **kwargs: Any) -> None:
        """
        Set model options.

        While model parameters are stored as a dictionary of string: float
        pairs, model options may be used to store other information such
        as switching between different model variants.

        Examples
        --------
        >>> from cymr import parameters
        >>> param_def = parameters.Parameters()
        >>> param_def.set_options(scope='list', recall_segment='item')
        >>> param_def.set_options({'option1': True, 'option2': False})
        """
        self.options.update(*args, **kwargs)

    def set_fixed(self, *args: dict[str, float], **kwargs: float):
        """
        Set fixed parameter values.

        Examples
        --------
        >>> from cymr import parameters
        >>> param_def = parameters.Parameters()
        >>> param_def.set_fixed(a=1, b=2)
        >>> param_def.set_fixed({'c': 3, 'd': 4})
        """
        self.fixed.update(*args, **kwargs)

    def set_free(
        self, *args: dict[str, Iterable[float]], **kwargs: Iterable[float]
    ) -> None:
        """
        Set free parameter ranges.

        Examples
        --------
        >>> from cymr import parameters
        >>> param_def = parameters.Parameters()
        >>> param_def.set_free(a=[0, 1], b=[1, 10])
        >>> param_def.set_free({'c': [3, 4], 'd': [0, 10]})
        """
        self.free.update(*args, **kwargs)

    def set_dependent(self, *args: dict[str, str], **kwargs: str) -> None:
        """
        Set dependent parameters in terms of other parameters.

        Examples
        --------
        >>> from cymr import parameters
        >>> param_def = parameters.Parameters()
        >>> param_def.set_dependent(a='exp(b * 2)')
        >>> param_def.set_dependent({'b': 'a * c', 'd': '2 * c + 1'})
        """
        self.dependent.update(*args, **kwargs)

    def eval_dependent(self, param: dict[str, float]) -> dict[str, float]:
        """
        Evaluate dependent parameters based on input parameters.

        Parameters
        ----------
        param : dict of str: float
            Parameters to use when evaluating dependent parameters.

        Returns
        -------
        eval_param : dict of str: float
            Input parameters with dependent parameters set.

        Examples
        --------
        >>> from cymr import parameters
        >>> param_def = parameters.Parameters()
        >>> param_def.set_dependent(b='clip(3 * a, 0, 1)')
        >>> param_def.eval_dependent({'a': 0.5})
        {'a': 0.5, 'b': 1.0}
        """
        return set_dependent(param, self.dependent)

    def set_dynamic(
        self, trial_type: str, scope: str, *args: dict[str, str], **kwargs: str
    ) -> None:
        """
        Set dynamic parameters in terms of parameters and data.

        Parameters
        ----------
        trial_type : str
            Type of trial that the parameter will vary over.
        
        scope : str
            Scope that the parameter applies to. May be "item" or 
            "list".

        Examples
        --------
        >>> from cymr import parameters
        >>> param_def = parameters.Parameters()
        >>> param_def.set_dynamic('study', a='b * input')
        >>> param_def.set_dynamic('recall', {'c': 'd * op'})
        """
        call_dict = dict(*args, **kwargs)
        if trial_type in self.dynamic:
            self.dynamic[trial_type].update(*args, **kwargs)
        else:
            self.dynamic[trial_type] = dict(*args, **kwargs)
        if 'scope' not in self.dynamic:
            self.dynamic['scope'] = {}
        for key in call_dict.keys():
            self._dynamic_names.add(key)
            self.dynamic['scope'][key] = scope

    def eval_dynamic(
        self,
        param: dict[str, Any],
        study: Optional[dict[str, list[ArrayLike]]] = None,
        recall: Optional[dict[str, list[ArrayLike]]] = None
    ) -> dict[str, Union[float, list[ArrayLike]]]:
        """
        Evaluate dynamic parameters based on data fields.

        Parameters
        ----------
        param : dict of str: float
            Parameters to use when evaluating dynamic parameters.

        study : dict of list of numpy.array, optional
            Study data to use when evaluating parameters.

        recall : dict of list of numpy.array, optional
            Recall data to use when evaluating parameters.

        Returns
        -------
        eval_param : dict of str: float
            Input parameters with dynamic parameters set.

        Examples
        --------
        >>> from cymr import parameters
        >>> param_def = parameters.Parameters()
        >>> param_def.set_dynamic('study', a='b * input')
        >>> data = {'input': [np.array([1, 2, 3])]}
        >>> param_def.eval_dynamic({'b': 0.2}, study=data)
        {'b': 0.2, 'a': [array([0.2, 0.4, 0.6])]}
        """
        if 'study' in self.dynamic and study is not None:
            param = set_dynamic(
                param, study, self.dynamic['study'], self.dynamic['scope']
            )
        if 'recall' in self.dynamic and recall is not None:
            param = set_dynamic(
                param, recall, self.dynamic['recall'], self.dynamic['scope']
            )
        return param

    def get_dynamic(
        self, param: dict[str, Any], index: int
    ) -> dict[str, Union[float, list[ArrayLike]]]:
        """
        Get list-specific parameters.

        Parameters
        ----------
        param : dict
            Parameters to index.

        index : int
            Index of the list.

        Returns
        -------
        indexed : dict
            Copy of input parameters with dynamic parameters for the
            specified list.
        """
        indexed = param.copy()
        for name in self._dynamic_names:
            indexed[name] = param[name][index]
        return indexed



"""Simulate free recall experiments."""

from importlib import resources
from abc import ABC, abstractmethod
import numpy as np
from scipy import optimize
import pandas as pd
from joblib import Parallel, delayed
from psifr import fr

from cymr import parameters


def sample_data(study):
    """Read sample data."""
    ref = resources.files("cymr").joinpath(f"data/{study}.csv")
    with resources.as_file(ref) as data_file:
        df = pd.read_csv(data_file)
    return df


def prepare_lists(data, study_keys=None, recall_keys=None, clean=True):
    """
    Prepare study and recall data for simulation.

    Return data information split by list. This format is similar to
    frdata structs used in EMBAM.

    Parameters
    ----------
    data : pandas.DataFrame
        Free recall data in Psifr format.

    study_keys : list of str, optional
        Columns to export for study list data. Default is:
        ['input', 'item_index']. Input position is assumed to be
        one-indexed.

    recall_keys : list of str, optional
        Columns to export for recall list data. Default is:
        ['input', 'item_index']. Input position is assumed to be
        one-indexed.

    clean : bool, optional
        If true, repeats and intrusions will be removed.

    Returns
    -------
    study : dict of (str: list of numpy.array)
        Study columns in list format.

    recall : dict of (str: list of numpy.array)
        Recall columns in list format.
    """
    if study_keys is None:
        study_keys = ['input', 'item_index']

    if recall_keys is None:
        recall_keys = ['input', 'item_index']

    s_keys = study_keys.copy()
    s_keys.remove('input')
    r_keys = recall_keys.copy()
    r_keys.remove('input')
    if 'item_index' in r_keys:
        r_keys.remove('item_index')
    merged = fr.merge_free_recall(data, study_keys=s_keys, recall_keys=r_keys)
    if clean:
        merged = merged.query('~intrusion and repeat == 0')

    study = fr.split_lists(merged, 'study', study_keys)
    recall = fr.split_lists(merged, 'recall', recall_keys)

    for i in range(len(study['input'])):
        if 'input' in study_keys:
            study['input'][i] = study['input'][i].astype(int) - 1
        if 'item_index' in study_keys:
            study['item_index'][i] = study['item_index'][i].astype(int)

        if 'input' in recall_keys:
            recall['input'][i] = recall['input'][i].astype(int) - 1
        if 'item_index' in recall_keys:
            recall['item_index'][i] = recall['item_index'][i].astype(int)

    n = np.unique([len(items) for items in study['input']])
    if len(n) > 1:
        raise ValueError('List length must not vary.')
    return study, recall


def prepare_study(study_data, study_keys=None):
    """
    Prepare study phase data for simulation.

    Parameters
    ----------
    study_data : pandas.DataFrame
        Study list data. Position is assumed to be one-indexed.

    study_keys : list of str
        Columns to export to split list format.

    Returns
    -------
    study : dict of (str: numpy.array)
        Study columns in split list format.
    """
    if study_keys is None:
        study_keys = ['position', 'item_index']

    study = fr.split_lists(study_data, 'raw', study_keys)
    for i in range(len(study['position'])):
        if 'position' in study_keys:
            study['position'][i] = study['position'][i].astype(int) - 1
        if 'item_index' in study_keys:
            study['item_index'][i] = study['item_index'][i].astype(int)
    return study


def add_recalls(study, recalls_list):
    """
    Add recall sequences to a study DataFrame.

    Parameters
    ----------
    study : pandas.DataFrame
        Study list data.

    recalls_list : list of numpy.array
        Recalled items for each list in output order.

    Returns
    -------
    data : pandas.DataFrame
        Complete free recall DataFrame suitable for analysis.
    """
    lists = study['list'].unique()
    subjects = study['subject'].unique()
    if len(subjects) > 1:
        raise ValueError('Unpacking multiple subjects not supported.')
    subject = subjects[0]

    # initialize recall trials DataFrame
    n_recall = np.sum([len(r) for r in recalls_list])
    recall = pd.DataFrame(
        {
            'subject': subject,
            'list': np.zeros(n_recall, dtype=int),
            'trial_type': 'recall',
            'position': np.zeros(n_recall, dtype=int),
            'item': '',
        }
    )

    # set basic information (list, item, position)
    n = 0
    for i, seq in enumerate(recalls_list):
        for j, item in enumerate(seq):
            recall.loc[n, 'list'] = lists[i]
            recall.loc[n, 'item'] = item
            recall.loc[n, 'position'] = j + 1
            n += 1
    data = pd.concat((study, recall), axis=0, ignore_index=True)
    data = data.sort_values(['list', 'trial_type'], ascending=[True, False])
    return data


def get_best_results(results):
    """Get best results from a repeated search."""
    df = []
    subjects = results.index.get_level_values('subject').unique()
    for subject in subjects:
        res = results.loc[subject].reset_index()
        subject_best = res.loc[[res['logl'].argmax()]]
        df.append(subject_best)
    best = pd.concat(df, axis=0)
    best.index = subjects
    best.index.rename('subject', inplace=True)
    return best


class Recall(ABC):
    """
    Base class for evaluating a model of free recall.

    Common Parameters
    -----------------
    study : pandas.DataFrame
        Study list information.

    recall : pandas.DataFrame
        Recall period information for each list.

    param : dict
        Model parameter values.

    patterns : dict
        May include keys: 'vector' and/or 'similarity'. Vectors are
        used to set distributed model representations. Similarity
        matrices are used to set item connections. Vector and
        similarity values are dicts of (feature: array) specifying
        an array for one or more named features, with an
        [items x units] array for vector representations, or
        [items x items] for similarity matrices.
    """

    @abstractmethod
    def prepare_sim(self, subject_data, study_keys=None, recall_keys=None):
        """
        Prepare data for simulation.

        Exporting data in DataFrame format to list format for
        simulation takes time, so this is only done once before running
        a parameter search.

        Parameters
        ----------
        subject_data : pandas.DataFrame
            Data for one subject.

        study_keys : list of str
            Data columns to include in the study data.

        recall_keys : list of str
            Data columns to include in the recall data.

        Returns
        -------
        study : dict of (str: list of numpy.array)
            Information about the study phase in list format.

        recall : dict of (str: list of numpy.array)
            Information about recalled items in list format.
        """
        pass

    def prepare_subject(
        self,
        subject,
        data,
        group_param,
        subj_param=None,
        param_def=None,
        study_keys=None,
        recall_keys=None,
    ):
        """
        Prepare parameters and data for a subject.

        Parameters
        ----------
        subject : str or int
            Identifier for the subject to be simulated.

        data : pandas.DataFrame
            Full dataset with all subjects.

        group_param : dict of (str: float)
            Values of parameters that apply to all subjects.

        subj_param : dict of (str: dict of (str: float)), optional
            Values of subject parameters, indexed by subject ID.

        param_def : cymr.parameters.Parameters, optional
            Parameter definition object specifying dependent and
            dynamic parameters.

        study_keys : list of str, optional
            Columns to export for study list data.

        recall_keys : list of str, optional
            Columns to export for recall list data.

        Returns
        -------
        study : dict of (str: list of numpy.array)
            Study columns in list format.

        recall : dict of (str: list of numpy.array)
            Recall columns in list format.

        param : dict of (str: float or numpy.array)
            Parameters with dependent and dynamic parameters evaluated.
        """
        param = group_param.copy()
        if subj_param is not None:
            param.update(subj_param[subject])

        # filter the data events for this subject
        subject_data = data.loc[data['subject'] == subject]

        # convert subject dataframe to list format
        study, recall = self.prepare_sim(
            subject_data, study_keys=study_keys, recall_keys=recall_keys
        )

        # evaluate dependent and dynamic parameters
        if param_def is not None:
            param = param_def.eval_dependent(param)
            param = param_def.eval_dynamic(param, study, recall)
        return study, recall, param

    @abstractmethod
    def likelihood_subject(self, study, recall, param, param_def=None, patterns=None):
        """
        Log likelihood of data for one subject based on a given model.

        Parameters
        ----------
        study : dict of (str: list of numpy.array)
            Information about the study phase in list format.

        recall : dict of (str: list of numpy.array)
            Information about recalled items in list format.

        param : dict of (str: float)
            Model parameter values.

        param_def : cymr.parameters.Parameters, optional
            Parameter definition object; used to interpret parameters.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        Returns
        -------
        logl : float
            Total log likelihood of data for this subject.

        n : int
            Number of evaluated data points.
        """
        pass

    def eval_model_stats(
        self, 
        data, 
        study, 
        recall, 
        param, 
        param_def, 
        patterns, 
        stats_def, 
        stats_data,
        n_stats_rep=1,
    ):
        """
        Simulate data and evaluate fit to summary statistics.

        Parameters
        ----------
        data : pandas.DataFrame
            Data for one subject.

        study : dict of (str: list of numpy.array)
            Information about the study phase in list format.

        recall : dict of (str: list of numpy.array)
            Information about recall trials in list format.

        param : dict of (str: float)
            Model parameter values.

        param_def : cymr.parameters.Parameters
            Parameter definitions.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        stats_def : cymr.statistics.Statistics
            Statistics to use when evaluating the model fit. If None,
            will evaluate based on likelihood.

        stats_data : pandas.DataFrame
            Summary statistics for the comparison data. Must have been
            evaluated using the same stat_def.
        
        n_stats_rep : int, optional
            Number of times to replicate generation and stat 
            evaluation.
        """
        stats_fit_rep = []
        for i in range(n_stats_rep):
            # generate simulated data
            recalls_list = self.generate_subject(study, recall, param, param_def, patterns)
            study_data = data[data['trial_type'] == 'study']
            full_data = add_recalls(study_data, recalls_list)

            # calculate summary stats
            merged = fr.merge_free_recall(full_data)
            stats_fit, _ = stats_def.eval_stats(merged)
            stats_fit_rep.append(stats_fit)

        # calculate mean over replications        
        reps = list(range(n_stats_rep))
        group_vars = ["stat", "point", "conditions", "subject", "independent"]
        stats_mean = (
            pd.concat(stats_fit_rep, axis=0, keys=reps, names=["rep", "point"])
            .groupby(group_vars)["dependent"]
            .mean()
            .reset_index()
        )

        # compare to data being fitted
        error_stat = stats_def.compare_stats(stats_data, stats_mean)
        return error_stat


    def likelihood(
        self,
        data,
        group_param,
        subj_param=None,
        param_def=None,
        patterns=None,
        study_keys=None,
        recall_keys=None,
    ):
        """
        Log likelihood summed over all subjects.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to fit. Must include a 'subject' column.

        group_param : dict of (str: float)
            Parameters that are fixed at the group level.

        subj_param : dict of (str: dict of (str: float))
            Parameters that vary by subject, indexed by subject.

        param_def : cymr.parameters.Parameters
            Parameter definitions (used to set dependent and dynamic).

        patterns : dict of (str: dict of (str: numpy.array))
            Patterns to use in the model.

        study_keys : list of str
            Fields to include in study data.

        recall_keys : list of str
            Fields to include in recall data.

        Returns
        -------
        stats : pandas.DataFrame
            Statistics related to model fitness. Includes logl, the
            log likelihood for each subject, and n, the number of data
            points for each subject.
        """
        # get the list of subjects
        subjects = data['subject'].unique()
        logl = np.empty(len(subjects))
        n = np.empty(len(subjects), int)
        for i, subject in enumerate(subjects):
            # prepare subject for simulation
            study, recall, param = self.prepare_subject(
                subject,
                data,
                group_param,
                subj_param,
                param_def,
                study_keys,
                recall_keys,
            )

            # run subject-specific likelihood function
            subject_logl, subject_n = self.likelihood_subject(
                study, recall, param, param_def=param_def, patterns=patterns
            )
            logl[i] = subject_logl
            n[i] = subject_n
        results = pd.DataFrame({'logl': logl, 'n': n}, index=subjects)
        results.index.rename('subject', inplace=True)
        return results

    def fit_subject(
        self, 
        subject_data, 
        param_def, 
        patterns=None, 
        study_keys=None, 
        recall_keys=None, 
        stats_def=None, 
        n_stats_rep=1,
        method='de', 
        **kwargs,
    ):
        """
        Fit a model to data for one subject.

        Parameters
        ----------
        subject_data : pandas.DataFrame
            Data for one subject.

        param_def : cymr.parameters.Parameters
            Parameter definitions.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        study_keys : list of str
            Fields to include in study data.

        recall_keys : list of str
            Fields to include in recall data.

        stats_def : cymr.statistics.Statistics
            Statistics to use when evaluating the model fit. If None,
            will evaluate based on likelihood.
        
        n_stats_rep : int, optional
            Number of times to replicate generation and stat 
            evaluation.

        method : str, optional
            Search method for fitting the parameters.

        kwargs
            Additional keyword arguments for the search method.

        Returns
        -------
        param : dict of (str: float)
            Best-fitting parameters, including fixed, free, and
            dependent parameters.

        logl : float
            Log likelihood for the best-fitting parameters.

        n : int
            Number of data points evaluated.

        k : int
            Number of free parameters.
        """
        study, recall = self.prepare_sim(subject_data, study_keys, recall_keys)
        var_names = list(param_def.free.keys())
        if stats_def is not None:
            # evaluate stats on data to fit
            subject_merged = fr.merge_free_recall(subject_data)
            stats_data, _ = stats_def.eval_stats(subject_merged)

        def eval_fit(x):
            eval_param = param_def.fixed.copy()
            eval_param.update(dict(zip(var_names, x)))
            eval_param = param_def.eval_dependent(eval_param)
            eval_param = param_def.eval_dynamic(eval_param, study, recall)
            
            if stats_def is not None:
                # generate data, calculate stats, compare to data stats
                error_stat = self.eval_model_stats(
                    subject_data, 
                    study, 
                    recall, 
                    eval_param, 
                    param_def, 
                    patterns, 
                    stats_def, 
                    stats_data,
                    n_stats_rep,
                )
            else:
                # calculate log likelihood of recall sequences
                eval_logl, _ = self.likelihood_subject(
                    study, recall, eval_param, param_def, patterns
                )
                if np.isnan(eval_logl):
                    eval_logl = -10e6
                error_stat = -eval_logl
            return error_stat

        group_lb = [param_def.free[k][0] for k in var_names]
        group_ub = [param_def.free[k][1] for k in var_names]
        bounds = optimize.Bounds(group_lb, group_ub)
        if method == 'de':
            res = optimize.differential_evolution(eval_fit, bounds, **kwargs)
        elif method == 'shgo':
            b = [(lb, ub) for lb, ub in zip(bounds.lb, bounds.ub)]
            res = optimize.shgo(eval_fit, b, **kwargs)
        else:
            raise ValueError(f'Invalid method: {method}')

        # get fitted parameters
        param = param_def.fixed.copy()
        param.update(dict(zip(var_names, res['x'])))
        param = param_def.eval_dependent(param)
        param_dynamic = param_def.eval_dynamic(param, study, recall)

        # evaluate fitted parameters, get number of fitted points
        if stats_def is not None:
            fit_stat = self.eval_model_stats(
                subject_data, 
                study, 
                recall, 
                param_dynamic, 
                param_def, 
                patterns, 
                stats_def, 
                stats_data,
                n_stats_rep,
            )
            n = sum(len(r) + 1 for r in recall['input'])
        else:
            logl, n = self.likelihood_subject(
                study, recall, param_dynamic, param_def, patterns
            )
            assert logl == -res['fun']
            fit_stat = logl
        k = len(param_def.free)
        return param, fit_stat, n, k

    def _run_fit_subject(
        self, 
        data, 
        subject, 
        param_def, 
        patterns=None, 
        study_keys=None, 
        recall_keys=None, 
        stats_def=None, 
        n_stats_rep=1,
        method='de', 
        **kwargs,
    ):
        """Apply fitting to one subject."""
        subject_data = data.loc[data['subject'] == subject]
        param, fit_stat, n, k = self.fit_subject(
            subject_data, 
            param_def, 
            patterns, 
            study_keys, 
            recall_keys, 
            stats_def, 
            n_stats_rep, 
            method, 
            **kwargs,
        )
        if stats_def is None:
            results = {**param, 'logl': fit_stat, 'n': n, 'k': k}
        else:
            results = {
                **param, stats_def.options['error_stat']: fit_stat, 'n': n, 'k': k
            }
        return results

    def fit_indiv(
        self,
        data,
        param_def,
        patterns=None,
        study_keys=None,
        recall_keys=None,
        stats_def=None, 
        n_stats_rep=1,
        n_jobs=None,
        method='de',
        n_rep=1,
        **kwargs,
    ):
        """
        Fit parameters to individual subjects.

        Parameters may be estimated either based on evaluating the
        overall likelihood of recall sequences in the observed data,
        or based on specific summary statistics. 
        
        To fit to summary statistics, simulated data must be generated 
        first so they can be analyzed. Note that simulated data are 
        stochastic, and any variability in summary statistics may 
        complicate the parameter search. To help stabalize summary 
        statistics, simulated data may optionally replicate the study 
        multiple times.

        Parameters
        ----------
        data : pandas.DataFrame
            Data for one or more subjects.

        param_def : cymr.parameters.Parameters
            Parameter definitions.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        study_keys : list of str
            Fields to include in study data.

        recall_keys : list of str
            Fields to include in recall data.

        stats_def : cymr.analysis.Statistics
            Statistics to use when evaluating the model fit. If None,
            the fit will be evaluated based on the overall likelihood
            of the observed recall sequences. If a stats_def is 
            specified, simulated data will be generated and used to 
            calculate the statistics, which will be compared to the 
            observed data.
        
        n_stats_rep : int, optional
            Number of times to replicate generation and stat 
            evaluation. Only used if stats_def is not None.

        n_jobs : int, optional
            Number of processes to use for fitting subjects in
            parallel.

        method : str, optional
            Search method for fitting the parameters.

        n_rep : int, optional
            Number of times to repeat each search.

        kwargs
            Additional keyword arguments for the search method.

        Returns
        -------
        results : pandas.DataFrame
            Best-fitting parameters, log likelihood (:code:`logl`)
            or error statistic specified by stats_def, number of data 
            points (:code:`n`), and number of free parameters 
            (:code:`k`) for each subject.
        """
        subjects = data['subject'].unique()
        full_subjects = np.repeat(subjects, n_rep)
        full_reps = np.tile(np.arange(n_rep), len(subjects))
        if n_jobs == 1:
            full_results = [
                self._run_fit_subject(
                    data, 
                    subject, 
                    param_def, 
                    patterns, 
                    study_keys, 
                    recall_keys, 
                    stats_def,
                    n_stats_rep,
                    method, 
                    **kwargs,
                )
                for subject in full_subjects
            ]
        else:
            full_results = Parallel(n_jobs=n_jobs)(
                delayed(self._run_fit_subject)(
                    data, 
                    subject, 
                    param_def, 
                    patterns, 
                    study_keys, 
                    recall_keys, 
                    stats_def,
                    n_stats_rep,
                    method, 
                    **kwargs,
                )
                for subject in full_subjects
            )
        d = {
            (subject, rep): res
            for subject, rep, res in zip(full_subjects, full_reps, full_results)
        }
        results = pd.DataFrame(d).T
        results.index.rename(['subject', 'rep'], inplace=True)
        results = results.astype({'n': int, 'k': int})
        return results

    @abstractmethod
    def generate_subject(
        self, study, recall, param, param_def=None, patterns=None, **kwargs
    ):
        """
        Generate simulated data for one subject.

        Parameters
        ----------
        study : dict of (str: list of numpy.array)
            Information about the study phase in list format.

        recall : dict of (str: list of numpy.array)
            Information about recall trials in list format.

        param : dict of (str: float)
            Model parameter values.

        param_def : cymr.parameters.Parameters, optional
            Parameter definitions.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        Returns
        -------
        recalls_list : list of numpy.array
            Recalled items for each simulated list.
        """
        pass

    def generate(
        self,
        data,
        group_param,
        subj_param=None,
        param_def=None,
        patterns=None,
        study_keys=None,
        recall_keys=None,
        n_rep=1,
    ):
        """
        Generate simulated data for all subjects.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to guide simulation. Must include a 'subject' column.
            May include dummy recall events if there is a dynamic
            recall parameter.

        group_param : dict of (str: float)
            Values of parameters that apply to all subjects.

        subj_param : dict of (str: dict of (str: float))
            Parameters that vary by subject, indexed by subject.

        param_def : cymr.parameters.Parameters, optional
            Parameter definitions.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        study_keys : list of str
            Data columns to include for the study phase.

        recall_keys : list of str
            Data columns to include for the recall phase.

        n_rep : int
            Number of times to repeat the simulation for each subject.

        Returns
        -------
        sim_data : pandas.DataFrame
            Simulated data for each subject.
        """
        # get the list of subjects
        subjects = data['subject'].unique()
        data_list = []
        for subject in subjects:
            # filter the data events for this subject
            subject_data = data.loc[data['subject'] == subject]

            # prepare data and parameters
            study, recall, param = self.prepare_subject(
                subject,
                subject_data,
                group_param,
                subj_param,
                param_def,
                study_keys,
                recall_keys,
            )

            max_list = subject_data['list'].max()
            # iterate over repetitions for this subject
            for i in range(n_rep):
                # run subject-specific generation function
                rep_recalls_list = self.generate_subject(
                    study, recall, param, param_def, patterns=patterns
                )
                rep_data = subject_data.copy()
                rep_data['list'] = i * max_list + subject_data['list']

                # strip off the dummy recall events
                rep_data = rep_data[rep_data['trial_type'] == 'study']
                rep_data = add_recalls(rep_data, rep_recalls_list)
                data_list.append(rep_data)
        sim_data = pd.concat(data_list, axis=0, ignore_index=True)
        return sim_data

    @abstractmethod
    def record_subject(
        self, study, recall, param, param_def=None, patterns=None, **kwargs
    ):
        """
        Record model state during simulation of data for one subject.

        Parameters
        ----------
        study : dict of (str: list of numpy.array)
            Information about the study phase in list format.

        recall : dict of (str: list of numpy.array)
            Information about recalled items in list format.

        param : dict of (str: float)
            Model parameter values.

        param_def : cymr.parameters.Parameters, optional
            Parameter definition object; used to interpret parameters.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        Returns
        -------
        study_state : list of lists
            Recorded state for each study trial.

        recall_state : list of lists
            Recorded state for each recall attempt.
        """

    def record(
        self,
        data,
        group_param,
        subj_param=None,
        param_def=None,
        patterns=None,
        study_keys=None,
        recall_keys=None,
        **kwargs,
    ):
        """
        Record model states during a simulation.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to guide simulation. Must include a 'subject' column.
            May include dummy recall events if there is a dynamic
            recall parameter.

        group_param : dict of (str: float)
            Values of parameters that apply to all subjects.

        subj_param : dict of (str: dict of (str: float))
            Parameters that vary by subject, indexed by subject.

        param_def : cymr.parameters.Parameters, optional
            Parameter definitions.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        study_keys : list of str
            Data columns to include for the study phase.

        recall_keys : list of str
            Data columns to include for the recall phase.

        Returns
        -------
        states : list
            List of model states.
        """
        # get the list of subjects
        subjects = data['subject'].unique()
        states = []
        for subject in subjects:
            # filter the data events for this subject
            subject_data = data.loc[data['subject'] == subject]

            # prepare data and parameters
            study, recall, param = self.prepare_subject(
                subject,
                subject_data,
                group_param,
                subj_param,
                param_def,
                study_keys,
                recall_keys,
            )

            # record study and recall states
            study_state, recall_state = self.record_subject(
                study, recall, param, param_def=param_def, patterns=patterns, **kwargs
            )

            # combine states into a flat list
            for study_list, recall_list in zip(study_state, recall_state):
                for trial_state in study_list:
                    states.append(trial_state)
                for trial_state in recall_list:
                    states.append(trial_state)
        return states

    def _run_parameter_recovery(
        self, data, param_def, patterns=None, method='de', n_rep=1, **kwargs
    ):
        """Run a parameter recovery test."""
        # generate parameters
        param = param_def.fixed.copy()
        sampled = parameters.sample_parameters(param_def.free)
        param.update(sampled)

        # generate simulated data
        sim = self.generate(
            data, param, None, param_def, patterns=patterns, n_rep=n_rep
        )

        # fit the simulated data
        fitted_param, logl, n, k = self.fit_subject(
            sim, param_def, patterns=patterns, method=method, **kwargs
        )

        # store results
        df_sim = pd.DataFrame(param, index=['sim'])
        df_fit = pd.DataFrame(fitted_param, index=['fit'])
        df_sample = pd.concat((df_sim, df_fit), axis=0)
        return df_sample

    def parameter_recovery(
        self,
        data,
        n_sample,
        param_def,
        patterns=None,
        method='de',
        n_rep=1,
        n_jobs=None,
        **kwargs,
    ):
        """Run multiple iterations of parameter recovery."""
        results_list = Parallel(n_jobs=n_jobs)(
            delayed(self._run_parameter_recovery)(
                data, param_def, patterns, method, n_rep, **kwargs
            )
            for i in range(n_sample)
        )
        results = pd.concat(results_list, axis=0, keys=np.arange(n_sample))
        return results

    def parameter_sweep(
        self,
        data,
        group_param,
        subj_param,
        sweep_param,
        param_def,
        patterns=None,
        n_rep=1,
    ):
        """
        Simulate data with varying parameters.

        Parameters
        ----------
        data : pandas.DataFrame
            Data to guide simulation. Must include a 'subject' column.
            May include dummy recall events if there is a dynamic
            recall parameter.

        group_param : dict of (str: float)
            Values of parameters that apply to all subjects.

        subj_param : dict of (str: dict of (str: float))
            Parameters that vary by subject, indexed by subject.

        sweep_param : dict of (str: numpy.ndarray)
            Parameter values to sweep over.

        param_def : cymr.parameters.Parameters, optional
            Parameter definitions.

        patterns : dict of (str: dict of (str: numpy.array)), optional
            Patterns to use in the model.

        n_rep : int
            Number of times to repeat the simulation for each subject.

        Returns
        -------
        results : pandas.DataFrame
            Simulated data for each combination of sweep parameters.
        """
        index = pd.MultiIndex.from_product(
            sweep_param.values(), names=sweep_param.keys()
        )
        df_list = []
        for idx in index:
            param = group_param.copy()
            param.update(dict(zip(sweep_param.keys(), idx)))
            sim = self.generate(
                data, param, subj_param, param_def, patterns=patterns, n_rep=n_rep
            )
            df_list.append(sim)
        results = pd.concat(df_list, axis=0, keys=index)
        results = results.droplevel(len(sweep_param.keys()))
        results.index.rename(sweep_param.keys(), inplace=True)
        return results



"""Test management of CMR parameters."""

import numpy as np
import pytest

from cymr import fit
from cymr import cmr


@pytest.fixture()
def param_def_simple():
    param = cmr.CMRParameters()
    param.set_sublayers(f=['task'], c=['task'])
    weights = {(('task', 'item'), ('task', 'item')): 'loc'}
    param.set_weights('fc', weights)
    param.set_weights('cf', weights)
    return param


def test_param_simple(param_def_simple):
    param_def = param_def_simple
    assert param_def.fixed == {}
    assert param_def.free == {}
    assert param_def.dependent == {}
    assert param_def.dynamic == {}
    assert param_def.sublayers == {'f': ['task'], 'c': ['task']}
    assert param_def.weights['fc'] == {(('task', 'item'), ('task', 'item')): 'loc'}
    assert param_def.weights['cf'] == {(('task', 'item'), ('task', 'item')): 'loc'}
    assert param_def.sublayer_param == {}


@pytest.fixture()
def param_def():
    """Parameter definitions."""
    param = cmr.CMRParameters()

    # options
    param.set_options(scope='pool')

    # network definition
    param.set_sublayers(f=['task'], c=['loc', 'cat'])
    weights = {
        (('task', 'item'), ('loc', 'item')): 'loc',
        (('task', 'item'), ('cat', 'item')): 'cat',
    }
    param.set_weights('fc', weights)
    param.set_weights('cf', weights)
    param.set_weights('ff', {('task', 'item'): 'loc + cat'})

    # sublayer-varying parameters
    param.set_sublayer_param('c', 'loc', {'B_enc': 'B_enc_loc'})
    param.set_sublayer_param('c', 'cat', {'B_enc': 'B_enc_cat'})
    return param


def test_param(param_def):
    """Test that parameter definitions are correct."""
    assert param_def.options['scope'] == 'pool'
    assert param_def.sublayers == {'f': ['task'], 'c': ['loc', 'cat']}
    assert param_def.weights['fc'] == {
        (('task', 'item'), ('loc', 'item')): 'loc',
        (('task', 'item'), ('cat', 'item')): 'cat',
    }
    assert param_def.weights['cf'] == {
        (('task', 'item'), ('loc', 'item')): 'loc',
        (('task', 'item'), ('cat', 'item')): 'cat',
    }
    assert param_def.weights['ff'] == {('task', 'item'): 'loc + cat'}
    assert param_def.sublayer_param['c'] == {
        'loc': {'B_enc': 'B_enc_loc'},
        'cat': {'B_enc': 'B_enc_cat'},
    }


@pytest.fixture()
def data():
    """Base test DataFrame."""
    data = fit.sample_data('sample1')
    data['list'] = data['subject']
    data['subject'] = 1
    return data


@pytest.fixture()
def patterns():
    cat = np.zeros((24, 3))
    cat[:8, 0] = 1
    cat[8:16, 1] = 1
    cat[16:, 2] = 1
    sim_cat = np.zeros((24, 24))
    sim_cat[:8, :8] = 1
    sim_cat[8:16, 8:16] = 1
    sim_cat[16:, 16:] = 1
    patterns = {
        'vector': {
            'loc': np.eye(24),
            'cat': cat,
        },
        'similarity': {
            'loc': np.eye(24),
            'cat': sim_cat,
        },
    }
    return patterns


def test_eval_weights(param_def, patterns):
    weights = param_def.eval_weights(patterns)
    np.testing.assert_array_equal(
        weights['fc'][(('task', 'item'), ('loc', 'item'))], patterns['vector']['loc']
    )
    np.testing.assert_array_equal(
        weights['fc'][(('task', 'item'), ('cat', 'item'))], patterns['vector']['cat']
    )
    np.testing.assert_array_equal(
        weights['cf'][(('task', 'item'), ('loc', 'item'))], patterns['vector']['loc']
    )
    np.testing.assert_array_equal(
        weights['cf'][(('task', 'item'), ('cat', 'item'))], patterns['vector']['cat']
    )
    np.testing.assert_array_equal(
        weights['ff'][('task', 'item')],
        patterns['similarity']['loc'] + patterns['similarity']['cat'],
    )


def test_eval_weights_index(param_def, patterns):
    item_index = np.arange(12)
    weights = param_def.eval_weights(patterns, item_index=item_index)
    assert weights['fc'][(('task', 'item'), ('loc', 'item'))].shape == (12, 24)
    assert weights['cf'][(('task', 'item'), ('loc', 'item'))].shape == (12, 24)
    assert weights['ff'][('task', 'item')].shape == (12, 12)


def test_eval_sublayer_param(param_def):
    param = {'B_enc_loc': 0.8, 'B_enc_cat': 0.4}
    eval_param = param_def.eval_sublayer_param('c', param, 3)
    expected = np.array([[0.8, 0.4], [0.8, 0.4], [0.8, 0.4]])
    np.testing.assert_array_equal(eval_param['B_enc'], expected)



"""Test management of parameters."""

import numpy as np
import pytest

from psifr import fr
from cymr import fit
from cymr import parameters


@pytest.fixture()
def param_def():
    """Parameter definitions."""
    param = parameters.Parameters()

    # general parameter management
    param.set_fixed(a=1, b=2)
    param.set_fixed({'c': 3})
    param.set_dependent(d='2 + mean([a, b])')
    param.set_dynamic('study', 'item', e='distract / c')
    param.set_dynamic('study', 'list', f='clip(input, 0, 1)')
    param.set_free(f=[0, 1])
    return param


def test_param(param_def):
    """Test that parameter definitions are correct."""
    assert param_def.fixed == {'a': 1, 'b': 2, 'c': 3}
    assert param_def.free == {'f': [0, 1]}
    assert param_def.dependent == {'d': '2 + mean([a, b])'}
    assert param_def.dynamic == {
        'study': {'e': 'distract / c', 'f': 'clip(input, 0, 1)'}, 
        'scope': {'e': 'item', 'f': 'list'},
    }


@pytest.fixture()
def data():
    """Base test DataFrame."""
    data = fit.sample_data('sample1')
    data['list'] = data['subject']
    data['subject'] = 1
    return data


@pytest.fixture()
def split_data(data):
    """Data split into study and recall."""
    merged = fr.merge_free_recall(data, study_keys=['distract'])
    split = {
        'study': fr.split_lists(merged, 'study', ['input', 'distract']),
        'recall': fr.split_lists(merged, 'recall', ['input']),
    }
    return split


def test_set_dependent():
    param = {'Lfc': 0.7}
    dependent = {'Dfc': '1 - Lfc'}
    updated = parameters.set_dependent(param, dependent)
    expected = {'Lfc': 0.7, 'Dfc': 0.3}
    np.testing.assert_allclose(updated['Dfc'], expected['Dfc'])


def test_set_dynamic(data):
    param = {'B_distract': 0.2}
    dynamic = {'study': {'B_enc': 'distract * B_distract'}}
    study_data = fr.filter_data(data, 1, 1, 'study')
    study = fr.split_lists(study_data, 'raw', ['distract'])
    updated = parameters.set_dynamic(param, study, dynamic['study'])
    expected = {'B_distract': 0.2, 'B_enc': [np.array([0.2, 0.4, 0.6])]}
    np.testing.assert_allclose(updated['B_enc'][0], expected['B_enc'][0])


def test_dependent(param_def):
    """Test evaluation of dependent parameters."""
    param = {'a': 1, 'b': 2}
    param = param_def.eval_dependent(param)
    assert param == {'a': 1, 'b': 2, 'd': 3.5}


def test_dynamic(param_def, split_data):
    """Test evaluation of dynamic parameters."""
    param = {'c': 2}
    param = param_def.eval_dynamic(param, study=split_data['study'])
    np.testing.assert_array_equal(param['e'][0], np.array([0.5, 1, 1.5]))
    np.testing.assert_array_equal(param['e'][1], np.array([1.5, 1, 0.5]))


def test_get_dynamic(param_def, split_data):
    """Test indexing of dynamic parameters."""
    param = {'c': 2}
    param = param_def.eval_dynamic(param, study=split_data['study'])

    param1 = param_def.get_dynamic(param, 0)
    np.testing.assert_array_equal(param1['e'], np.array([0.5, 1, 1.5]))
    param2 = param_def.get_dynamic(param, 1)
    np.testing.assert_array_equal(param2['e'], np.array([1.5, 1, 0.5]))


def test_blank(split_data):
    """Test that unspecified evaluation does nothing."""
    param_def = parameters.Parameters()
    param = {'a': 1, 'b': 2, 'c': 3}
    orig = param.copy()
    param = param_def.eval_dependent(param)
    param = param_def.eval_dynamic(
        param, study=split_data['study'], recall=split_data['recall']
    )
    assert param == orig


def test_json(param_def, tmp_path):
    p = tmp_path / 'parameters.json'
    param_def.to_json(p.as_posix())
    param = parameters.read_json(p.as_posix())
    assert param.options == param_def.options
    assert param.fixed == param_def.fixed
    assert param.free == param_def.free
    assert param.dependent == param_def.dependent
    assert param.dynamic == param_def.dynamic



"""Models of free recall."""

from __future__ import annotations
import json
import numpy as np
import h5py
from typing import Union, Any, Iterable, Optional, Tuple
from numpy.typing import ArrayLike
from cymr.fit import Recall
from cymr import fit
from cymr import network
from cymr.parameters import Parameters


Region = Union[Tuple[str, str], Tuple[Tuple[str, str], Tuple[str, str]]]


def save_patterns(h5_file, items, **kwargs):
    """
    Write patterns and similarity matrices to hdf5.

    Parameters
    ----------
    h5_file : str
        Path to hdf5 file to save patterns in.

    items : list of str
        Item strings corresponding to the patterns.

    Additional keyword arguments set named feature vectors. Feature
    vector arrays must have shape [items x units].
    """
    with h5py.File(h5_file, 'w') as f:
        # items
        dt = h5py.special_dtype(vlen=str)
        items = np.asarray(items)
        dset = f.create_dataset('items', items.shape, dtype=dt)
        for i, item in enumerate(items):
            dset[i] = item

        # features
        features = list(kwargs.keys())
        dset = f.create_dataset('features', (len(features),), dtype=dt)
        for i, feature in enumerate(features):
            dset[i] = feature

        # patterns
        for name, vectors in kwargs.items():
            # save vectors
            f.create_dataset('vector/' + name, data=vectors)

            # set pattern similarity to dot product
            sim = np.dot(vectors, vectors.T)
            f.create_dataset('similarity/' + name, data=sim)


def load_patterns(h5_file, features=None):
    """
    Load weights from an hdf5 file.

    Parameters
    ----------
    h5_file : str
        Path to file saved with `save_patterns`.

    features : list of str, optional
        Names of features to load. Default is to load all features.

    Returns
    -------
    patterns : dict of (str: dict of (str: numpy.array))
        Loaded patterns. The "vector" field contains vector patterns.
        The "similarity" field contains pairwise similarity matrices.
        Each type of pattern contains a field for each loaded feature.
    """
    with h5py.File(h5_file, 'r') as f:
        patterns = {
            'items': np.array([item for item in f['items'].asstr()]),
            'vector': {},
            'similarity': {},
        }

        if features is None:
            features = f['features'].asstr()

        for name in features:
            patterns['vector'][name] = f['vector/' + name][()]
            patterns['similarity'][name] = f['similarity/' + name][()]
    return patterns


def encode_region(region: Region) -> str:
    """Encode a region as a string."""
    if len(region) == 0:
        raise ValueError('Cannot encode an empty region.')
    elif isinstance(region[0], str):
        region_str = '-'.join(region)
    else:
        region_str = '_'.join('-'.join(segment) for segment in region)
    return region_str


def decode_region(
    region_str: str
) -> Union[tuple[str, ...], tuple[tuple[str, ...], ...]]:
    """Decode a region string."""
    if '_' in region_str:
        region = tuple([tuple(s.split('-')) for s in region_str.split('_')])
    else:
        region = tuple(region_str.split('-'))
    return region


def read_config(json_file):
    """Read model configuration from a JSON file."""
    with open(json_file, 'r') as f:
        par_dict = json.load(f)

    par = CMRParameters()
    if 'options' in par_dict:
        par.set_options(par_dict['options'])
    par.set_free(par_dict['free'])
    par.set_fixed(par_dict['fixed'])
    par.set_dependent(par_dict['dependent'])
    for trial_type, p in par_dict['dynamic'].items():
        for name, val in p.items():
            if 'scope' in par_dict['dynamic'] and name in par_dict['dynamic']['scope']:
                scope = par_dict['dynamic']['scope'][name]
            else:
                scope = 'item'
            par.set_dynamic(trial_type, scope, {name: val})
    par.set_sublayers(par_dict['sublayers'])
    for connect, p in par_dict['weights'].items():
        weight_dict = {decode_region(region): expr for region, expr in p.items()}
        par.set_weights(connect, weight_dict)
    for layer, sublayer_param in par_dict['sublayer_param'].items():
        for sublayer, param in sublayer_param.items():
            par.set_sublayer_param(layer, sublayer, param)
    return par


class CMRParameters(Parameters):
    """
    Configuration of CMR model parameters.

    Attributes
    ----------
    fixed : dict of (str: float)
        Values of fixed parameters.

    free : dict of (str: tuple)
        Bounds of each free parameter.

    dependent : dict of (str: str)
        Expressions to define dependent parameters based the other
        parameters.

    dynamic : dict of (str: dict of (str: str))
        First dict specifies trial_type for dynamic parameters,
        second dict keys are parameter names, and values are
        expressions specifying how to update the parameter.

    sublayers : dict of (list of str)
        Names of sublayers for each layer in the network.

    weights : dict of (tuple of (tuple of str)): str
        Weights template to set network connections. Weights are
        indicated by region within the network. Each region is
        specified with a tuple giving names of sublayers and segments:
        ((f_sublayer, f_segment), (c_sublayer, c_segment)). The value
        for each region should be an expression to be evaluated with
        patterns and/or parameters.

    sublayer_param : dict of (str: dict of (str: dict of str))
        Parameters that vary by sublayer. These parameters are
        specified in terms of their layer and sublayer. Each value
        should contain an expression to be evaluated with parameters.
    """

    def __init__(self) -> None:
        super().__init__()
        self.options: dict[str, Any] = {
            'distraction': False, 'scope': 'list', 'filter_recalls': False
        }
        self.fixed: dict[str, float] = {}
        self.free: dict[str, Iterable[float]] = {}
        self.dependent: dict[str, str] = {}
        self.dynamic: dict[str, dict[str, str]] = {}
        self.sublayers: dict[str, list[str]] = {}
        self.weights: dict[str, dict[Region, str]] = {}
        self.sublayer_param: dict[str, dict[str, dict[str, str]]] = {}
        self._fields.extend(['options', 'sublayers', 'weights', 'sublayer_param'])

    def to_json(self, json_file: str) -> None:
        """
        Write parameter definitions to a JSON file.

        Parameters
        ----------
        json_file : str
            Path to file to save json data.
        """
        data: dict[str, Any] = {
            'options': self.options,
            'fixed': self.fixed,
            'free': self.free,
            'dependent': self.dependent,
            'dynamic': self.dynamic,
            'sublayers': self.sublayers,
            'weights': {},
            'sublayer_param': self.sublayer_param,
        }
        for layer, regions in self.weights.items():
            data['weights'][layer] = {}
            for region, expr in regions.items():
                region_str = encode_region(region)
                data['weights'][layer][region_str] = expr
        with open(json_file, 'w') as f:
            json.dump(data, f, indent=4)

    def set_sublayers(self, *args: dict[str, list[str]], **kwargs: list[str]) -> None:
        """
        Set layers and sublayers of a network.

        Examples
        --------
        >>> from cymr.cmr import CMRParameters
        >>> param_def = CMRParameters()
        >>> param_def.set_sublayers(f=['task'], c=['task'])
        """
        self.sublayers.update(*args, **kwargs)

    def set_weights(
        self, connect: str, regions: dict[Region, str]
    ) -> None:
        """
        Set weights on model patterns.

        Parameters
        ----------
        connect : str
            Network connection to set weights for.

        regions : dict of (tuple of (tuple of str)): str
            Weights for each region to set.

        Examples
        --------
        >>> from cymr.cmr import CMRParameters
        >>> param_def = CMRParameters()
        >>> region = (('f_sub', 'f_reg'), ('c_sub', 'c_reg'))
        >>> param_def.set_weights('fc', {region: 'b * pattern'})
        """
        if connect in self.weights:
            self.weights[connect].update(regions)
        else:
            self.weights[connect] = regions

    def eval_weights(
        self,
        patterns: dict[str, dict[str, Any]],
        param: Optional[dict[str, float]] = None,
        item_index: Optional[Any] = None
    ) -> dict[str, dict[Region, ArrayLike]]:
        """
        Evaluate weights based on parameters and patterns.

        Parameters
        ----------
        patterns : dict of str: (dict of str: numpy.ndarray)
            Patterns to use when evaluating weights.

        param : dict, optional
            Parameters to use when evaluating weights.

        item_index : numpy.ndarray, optional
            Item indices to include in the patterns.

        Returns
        -------
        weights : dict of str: (dict of str: numpy.ndarray)
            Weight matrices for each region in each connection matrix.
        """
        weights: dict[str, dict[Region, ArrayLike]] = {}
        for connect, regions in self.weights.items():
            # get necessary patterns
            weights[connect] = {}
            if connect in ['fc', 'cf']:
                layer_type = 'vector'
            elif connect == 'ff':
                layer_type = 'similarity'
            else:
                raise ValueError(f'Invalid connection: {connect}.')
            data = patterns[layer_type].copy()

            # filter by item index
            if item_index is not None:
                for feature, mat in data.items():
                    if layer_type == 'vector':
                        data[feature] = mat[item_index, :]
                    else:
                        data[feature] = mat[np.ix_(item_index, item_index)]

            # evaluate expressions to get weights
            if param is not None:
                data.update(param)
            for region, expr in regions.items():
                weights[connect][region] = eval(expr, np.__dict__, data)
        return weights

    def set_sublayer_param(
        self, layer: str, sublayer: str, *args: dict[str, str], **kwargs: str
    ) -> None:
        """
        Set sublayer parameters.

        Parameters
        ----------
        layer : str
            Layer containing the sublayer to set.

        sublayer : str
            Sublayer to set parameters for.

        Examples
        --------
        >>> from cymr.cmr import CMRParameters
        >>> param_def = CMRParameters()
        >>> param_def.set_sublayer_param('c', 'sub1', a='2 * d'})
        >>> param_def.set_sublayer_param('c', 'sub1', {'b': 'exp(f)'})
        """
        if layer in self.sublayer_param:
            if sublayer in self.sublayer_param[layer]:
                self.sublayer_param[layer][sublayer].update(*args, **kwargs)
            else:
                self.sublayer_param[layer][sublayer] = dict(*args, **kwargs)
        else:
            self.sublayer_param[layer] = {sublayer: dict(*args, **kwargs)}

    def eval_sublayer_param(
        self, layer: str, param: dict[str, Any], n_trial: int = None
    ) -> dict[str, Union[float, ArrayLike]]:
        """
        Evaluate sublayer parameters.

        Parameters
        ----------
        layer : str
            Layer to evaluate.

        param : dict
            Parameters to use when evaluating sublayer parameters.

        n_trial : int, optional
            Number of trials. If indicated, parameters will be tiled
            over all trials.

        Returns
        -------
        eval_param : dict
            Input parameters with sublayer-specific parameters set.
        """
        eval_param = param.copy()

        # get parameter values for each sublayer
        param_lists: dict[str, list[ArrayLike]] = {}
        for sublayer in self.sublayers[layer]:
            for par, expr in self.sublayer_param[layer][sublayer].items():
                if par not in param_lists:
                    param_lists[par] = []
                value = eval(expr, np.__dict__, param)
                param_lists[par].append(value)

        # prepare parameter arrays
        for par, values in param_lists.items():
            if any([isinstance(v, np.ndarray) for v in values]):
                size = (max([v.size for v in values if isinstance(v, np.ndarray)]),)
                eval_param[par] = np.asarray(
                    [network.expand_param(v, size) for v in values]
                ).T
            elif n_trial is not None:
                eval_param[par] = np.tile(np.asarray(values), (n_trial, 1))
            else:
                eval_param[par] = np.array(values)
        return eval_param


def primacy(n_item, L, P1, P2):
    """
    Primacy gradient in learning rate.

    Parameters
    ----------
    n_item : int
        Number of items in study list.

    L : float
        Base learning rate. Asymptote of gradient for later positions.

    P1 : float
        Additional learning for first item.

    P2 : float
        Decay rate for primacy gradient.

    Returns
    -------
    rate : numpy.array
        Learning rate for each serial position.
    """
    position = np.arange(n_item)
    rate = L + (P1 * np.exp(-P2 * position))
    return rate


def p_stop_op(n_item, X1, X2, pmin=0.000001):
    """
    Probability of stopping based on output position.

    Parameters
    ----------
    n_item : int
        Number of items available for recall.

    X1 : float
        Probability of not recalling any items.

    X2 : float
        Shape parameter of exponential function increasing stop
        probability by output position.

    pmin : float, optional
        Minimum probability of stopping recall at any position.

    Returns
    -------
    p_stop : numpy.array
        Probability of stopping for each output position.
    """
    p_stop = X1 * np.exp(X2 * np.arange(n_item + 1))
    p_stop[p_stop < pmin] = pmin
    p_stop[p_stop > (1 - pmin)] = 1 - pmin

    # after recalling all items, P(stop)=1 by definition
    p_stop[-1] = 1
    return p_stop


def config_loc_cmr(n_item):
    """
    Configure a localist CMR network.

    Parameters
    ----------
    n_item : int
        Number of item patterns to include in the network

    Returns
    -------
    param_def : cymr.parameters.Parameters
        Parameters object with configuration for the network.

    patterns : dict
        Patterns to place in the network.
    """
    items = np.arange(n_item)
    patterns = {'items': items, 'vector': {'loc': np.eye(n_item)}}
    param_def = CMRParameters()
    param_def.set_sublayers(f=['task'], c=['task'])
    param_def.set_weights(
        'fc', {(('task', 'item'), ('task', 'item')): 'Afc + Dfc * loc'}
    )
    param_def.set_weights(
        'cf', {(('task', 'item'), ('task', 'item')): 'Acf + Dcf * loc'}
    )
    return param_def, patterns


def init_network(param_def, patterns, param, item_index, remove_blank=False):
    """
    Initialize a network with pattern weights.

    Parameters
    ----------
    param_def : cymr.cmr.CMRParameters
        Parameters definition defining network sublayers and weights.

    patterns : dict
        Patterns to place in the network.

    param : dict of (str: float)
        Parameter values; used to initialize network weights.

    item_index : numpy.array
        Indices of item patterns to include in the network.

    remove_blank : bool
        If true, context units with zero weights will be removed.

    Returns
    -------
    net : cymr.network.Network
        Network initialized with pre-experimental weights.
    """
    # set item weights
    weights = param_def.eval_weights(patterns, param, item_index)

    if remove_blank:
        # remove context units that are zero for all items
        for connect in ['fc', 'cf']:
            for region, mat in weights[connect].items():
                include = np.any(mat != 0, 0)
                if not any(include):
                    # if no features nonzero, leave them all in
                    continue
                weights[connect][region] = mat[:, include]

    # set task units
    for f_sublayer in param_def.sublayers['f']:
        for c_sublayer in param_def.sublayers['c']:
            region = ((f_sublayer, 'start'), (c_sublayer, 'start'))
            weights['fc'][region] = np.array([[1]])
            weights['cf'][region] = np.array([[1]])

    if param_def.options['distraction']:
        # add distraction units
        n_distract = len(item_index) + 1
        for f_sublayer in param_def.sublayers['f']:
            for c_sublayer in param_def.sublayers['c']:
                region = ((f_sublayer, 'distract'), (c_sublayer, 'distract'))
                weights['fc'][region] = np.eye(n_distract)
                weights['cf'][region] = np.eye(n_distract)

    # get all segment definitions
    f_segments = {}
    c_segments = {}
    for region, mat in weights['fc'].items():
        f_region, c_region = region
        f_sublayer, f_segment = f_region
        c_sublayer, c_segment = c_region
        if f_sublayer not in f_segments:
            f_segments[f_sublayer] = {}
        if c_sublayer not in c_segments:
            c_segments[c_sublayer] = {}
        f_segments[f_sublayer][f_segment] = mat.shape[0]
        c_segments[c_sublayer][c_segment] = mat.shape[1]

    # initialize the network
    net = network.Network(f_segments, c_segments)
    for connect in weights.keys():
        for region, mat in weights[connect].items():
            if connect == 'ff':
                f_segment = region
                c_segment = None
            else:
                f_segment, c_segment = region
            net.add_pre_weights(connect, f_segment, c_segment, mat)
    return net


def study_list(param_def, param, item_index, item_input, patterns, item_distract):
    """
    Simulate study of a list.

    Parameters
    ----------
    param_def : cymr.cmr.CMRParameters
        Parameters definition defining network sublayers and weights.

    param : dict of (str: float)
        Parameter values; used to set context evolution and learning.

    item_index : numpy.array
        Indices of presented item patterns.

    item_input : numpy.array
        Input position of presented items.

    patterns : dict
        Item pattern vectors.
    
    item_distract : numpy.array
        Indices of distraction items.

    Returns
    -------
    net : cymr.network.Network
        Network after simulation of the study phase.
    """
    net = init_network(param_def, patterns, param, item_index)
    net.update(('task', 'start', 0), net.c_sublayers)
    if param_def.options['distraction']:
        net.study_distract(
            ('task', 'item'),
            item_input,
            net.c_sublayers,
            param['B_enc'],
            param['Lfc'],
            param['Lcf'],
            ('task', 'distract'),
            item_distract,
            param['B_distract'],
            param['B_retention'],
        )
    else:
        net.study(
            ('task', 'item'),
            item_input,
            net.c_sublayers,
            param['B_enc'],
            param['Lfc'],
            param['Lcf'],
        )
    net.integrate(('task', 'start', 0), net.c_sublayers, param['B_start'])
    return net


def prepare_list_param(n_item, n_sub, param, param_def):
    """
    Prepare parameters that vary within list.

    Parameters
    ----------
    n_item : int
        Number of items in the list.

    n_sub : int
        Number of sublayers of context.

    param : dict of (str: float)
        Parameter values.

    param_def : cymr.cmr.CMRParameters
        Parameter definitions indicating sublayer parameters.

    Returns
    -------
    list_param : dict
        Parameters prepared for changes between sublayer and trials.
    """
    if 'c' in param_def.sublayer_param:
        # evaluate sublayer parameters
        param = param_def.eval_sublayer_param('c', param, n_item)

        # get the set of all sublayer parameters
        sub_param = set()
        for sublayer in param_def.sublayers['c']:
            for par in param_def.sublayer_param['c'][sublayer].keys():
                sub_param.add(par)
    else:
        sub_param = set()

    if 'Lfc' in sub_param:
        Lfc = param['Lfc']
    else:
        Lfc = np.tile(param['Lfc'], (n_item, n_sub)).astype(float)

    # apply primacy gradient to Lcf
    if 'Lcf' in sub_param:
        n_sub = param['Lcf'].shape[1]
        Lcf = np.zeros(param['Lcf'].shape)
        for i in range(n_sub):
            Lcf[:, i] = primacy(n_item, param['Lcf'][0, i], param['P1'], param['P2'])
    else:
        Lcf_trial = primacy(n_item, param['Lcf'], param['P1'], param['P2'])
        Lcf = np.tile(Lcf_trial[:, None], (1, n_sub))

    p_stop = p_stop_op(n_item, param['X1'], param['X2'])
    list_param = param.copy()
    list_param.update({'Lfc': Lfc, 'Lcf': Lcf, 'p_stop': p_stop})
    return list_param


def get_list_items(item_index, study, recall, list_ind, scope):
    """Get item units to present given the paradigm."""
    if scope == 'list':
        item_pool = study['item_index'][list_ind]
        item_study = study['input'][list_ind]
        item_recall = recall['input'][list_ind]
        item_distract = np.arange(len(item_study) + 1)
    elif scope == 'pool':
        item_pool = item_index
        item_study = study['item_index'][list_ind]
        item_recall = recall['item_index'][list_ind]
        item_distract = np.arange(len(item_study) + 1)
    else:
        raise ValueError(f'Invalid scope: {scope}')
    return item_pool, item_study, item_recall, item_distract


class CMR(Recall):
    """
    Context Maintenance and Retrieval model.

    **Model Parameters**

    Lfc : float
        Learning rate of item-context weights.

    Lcf : float
        Learning rate of context-item weights.

    P1 : float
        Additional context-item learning for first item.

    P2 : float
        Decay rate for primacy learning rate gradient.

    B_enc : float
        Integration rate during encoding.
    
    B_distract : float
        Integration rate of distraction units before each item 
        presentation. Only used if the "distraction" option is set
        to True.
    
    B_retention : float
        Integration rate of distraction units after the last
        item presentation. Only used if the "distraction" option is set
        to True.

    B_start : float
        Integration rate of start context reinstatement.

    B_rec : float
        Integration rate during recall.

    X1 : float
        Probability of not recalling any items.

    X2 : float
        Shape parameter of exponential function increasing stop
        probability by output position.

    **Parameter definition objects**

    Parameters objects are used to indicate sublayers to include
    in the network and to indicate how network weights should be
    initialized. The :code:`sublayers` and :code:`weights` attributes
    must be set.

    Parameters objects may also be used to define parameters that depend
    on other parameters and/or dynamic parameters that depend on columns
    of the input data.

    Finally, Parameters objects are used to define searches, using the
    :code:`fixed` and :code:`free` attributes.

    **Model Patterns**

    Patterns are used to define connections between the item and
    context layers and direct connections between items. Connections
    may be orthonormal as in many published variants of CMR, or they
    may be distributed, overlapping patterns.

    Patterns may include :code:`'vector'` and/or :code:`'similarity'` matrices.
    Vector representations are used to set the :math:`M^{FC}_{pre}`
    and :math:`M^{CF}_{pre}` matrices, while similarity matrices are
    used to set the :math:`M^{FF}_{pre}` matrix.

    Vector and similarity values are dicts of (feature: array) specifying
    an array for one or more named features, with an
    [items x units] array for vector representations, or
    [items x items] for similarity matrices.
    """

    def prepare_sim(self, data, study_keys=None, recall_keys=None):
        study_base = ['input', 'item_index']
        if study_keys is None:
            study_keys = study_base
        else:
            # only add the base key if it isn't already in there
            for term in study_base:
                if term not in study_keys:
                    study_keys += [term]
        recall_base = ['input', 'item_index']
        if recall_keys is None:
            recall_keys = recall_base
        else:
            for term in recall_base:
                if term not in recall_keys:
                    recall_keys += [term]

        study, recall = fit.prepare_lists(
            data, study_keys=study_keys, recall_keys=recall_keys, clean=True
        )
        return study, recall

    def likelihood_subject(self, study, recall, param, param_def=None, patterns=None):
        n_item = len(study['input'][0])
        n_list = len(study['input'])
        if param_def is None:
            raise ValueError('Must provide a Parameters object.')
        n_sub = len(param_def.sublayers['c'])

        item_index = np.arange(len(patterns['items']))
        logl = 0
        n = 0
        for i in range(n_list):
            # access the dynamic parameters needed for this list
            list_param = param.copy()
            list_param = param_def.get_dynamic(list_param, i)
            list_param = param_def.eval_dependent(list_param)
            list_param = prepare_list_param(n_item, n_sub, list_param, param_def)

            # simulate study
            item_pool, item_study, item_recall, item_distract = get_list_items(
                item_index, study, recall, i, param_def.options['scope']
            )
            net = study_list(
                param_def, list_param, item_pool, item_study, patterns, item_distract
            )

            # get recall probabilities
            p = net.p_recall(
                ('task', 'item'),
                item_recall,
                net.c_sublayers,
                list_param['B_rec'],
                list_param['T'],
                list_param['p_stop'],
            )
            n += p.size
            if np.any(np.isnan(p)):
                # if any response probability is undefined, so is the
                # overall likelihood
                logl = np.nan
                break

            # ensure probabilities are not zero before log transform
            p = np.clip(p, 10e-6, 1)
            logl += np.sum(np.log(p))
        return logl, n

    def generate_subject(
        self, study, recall, param, param_def=None, patterns=None, **kwargs
    ):
        n_item = len(study['input'][0])
        n_list = len(study['input'])
        if param_def is None:
            raise ValueError('Must provide a Parameters object.')
        n_sub = len(param_def.sublayers['c'])

        item_index = np.arange(len(patterns['items']))
        recalls_list = []
        for i in range(n_list):
            # access the dynamic parameters needed for this list
            list_param = param.copy()
            list_param = param_def.get_dynamic(list_param, i)
            list_param = param_def.eval_dependent(list_param)
            list_param = prepare_list_param(n_item, n_sub, list_param, param_def)

            # simulate study
            item_pool, item_study, item_recall, item_distract = get_list_items(
                item_index, study, recall, i, param_def.options['scope']
            )
            net = study_list(
                param_def, list_param, item_pool, item_study, patterns, item_distract
            )

            # simulate recall
            if param_def.options['filter_recalls']:
                recall_index = net.generate_recall(
                    ('task', 'item'),
                    net.c_sublayers,
                    list_param['B_rec'],
                    list_param['T'],
                    list_param['p_stop'],
                    filter_recalls=True,
                    A1=list_param['A1'],
                    A2=list_param['A2'],
                )
            else:
                recall_index = net.generate_recall(
                    ('task', 'item'),
                    net.c_sublayers,
                    list_param['B_rec'],
                    list_param['T'],
                    list_param['p_stop'],
                )

            items = patterns['items'][item_pool]
            recall_items = items[recall_index]
            recalls_list.append(recall_items)
        return recalls_list

    def record_subject(
        self,
        study,
        recall,
        param,
        param_def=None,
        patterns=None,
        remove_blank=False,
        include=None,
        exclude=None,
    ):
        n_item = len(study['input'][0])
        n_list = len(study['input'])
        if param_def is None:
            raise ValueError('Must provide a Parameters object.')
        n_sub = len(param_def.sublayers['c'])

        study_state = []
        recall_state = []
        item_index = np.arange(len(patterns['items']))
        for i in range(n_list):
            # access the dynamic parameters needed for this list
            list_param = param.copy()
            list_param = param_def.get_dynamic(list_param, i)
            list_param = param_def.eval_dependent(list_param)
            list_param = prepare_list_param(n_item, n_sub, list_param, param_def)

            # initialize the network
            net = init_network(
                param_def,
                patterns,
                list_param,
                study['item_index'][i],
                remove_blank=remove_blank,
            )
            net.update(('task', 'start', 0), net.c_sublayers)

            # record study phase
            item_pool, item_study, item_recall, item_distract = get_list_items(
                item_index, study, recall, i, param_def.options['scope']
            )
            if param_def.options['distraction']:
                distract_segment = ('task', 'distract')
                distract_B = list_param['B_distract']
                retention_B = list_param['B_retention']
            else:
                distract_segment = None
                distract_B = None
                retention_B = None

            list_study_state = net.record_study(
                ('task', 'item'),
                item_study,
                net.c_sublayers,
                list_param['B_enc'],
                list_param['Lfc'],
                list_param['Lcf'],
                distract_segment=distract_segment,
                distract_list=item_distract,
                distract_B=distract_B,
                retention_B=retention_B,
                include=include,
                exclude=exclude,
            )
            net.integrate(('task', 'start', 0), net.c_sublayers, list_param['B_start'])

            # record recall phase
            list_recall_state = net.record_recall(
                ('task', 'item'),
                recall['input'][i],
                net.c_sublayers,
                list_param['B_rec'],
                list_param['T'],
                include=include,
                exclude=exclude,
            )
            study_state.append(list_study_state)
            recall_state.append(list_recall_state)
        return study_state, recall_state



"""Test fitting to summary statistics."""

import importlib
import pytest
import numpy as np
import pandas as pd
from psifr import fr
from cymr import analysis


@pytest.fixture()
def raw1():
    """Create raw free recall data."""
    subjects = [1, 1, 2, 2]
    lists = [1, 2, 1, 2]
    study = [
        ['absence', 'hollow', 'pupil'], 
        ['fountain', 'piano', 'pillow'],
        ['absence', 'hollow', 'pupil'], 
        ['fountain', 'piano', 'pillow'],
    ]
    recall = [
        ['hollow', 'pupil', 'empty'], 
        ['pillow', 'fountain', 'pillow'],
        ['absence', 'hollow'],
        ['piano', 'fountain', 'pillow'],
    ]
    item_index = ([[0, 1, 2], [3, 4, 5], [0, 1, 2], [3, 4, 5]], None)
    task = ([[1, 1, 1], [2, 2, 2], [1, 1, 1], [2, 2, 2]], None)
    raw = fr.table_from_lists(
        subjects, study, recall, lists, item_index=item_index, task=task
    )
    return raw


@pytest.fixture()
def raw2():
    """Create raw free recall data."""
    subjects = [1, 1, 2, 2]
    lists = [1, 2, 1, 2]
    study = [
        ['absence', 'hollow', 'pupil'], 
        ['fountain', 'piano', 'pillow'],
        ['absence', 'hollow', 'pupil'], 
        ['fountain', 'piano', 'pillow'],
    ]
    recall = [
        ['pupil', 'hollow'], 
        ['piano', 'fountain'],
        ['absence', 'pupil', 'absence'],
        ['fountain', 'piano', 'pillow'],
    ]
    item_index = ([[0, 1, 2], [3, 4, 5], [0, 1, 2], [3, 4, 5]], None)
    task = ([[1, 1, 1], [2, 2, 2], [1, 1, 1], [2, 2, 2]], None)
    raw = fr.table_from_lists(
        subjects, study, recall, lists, item_index=item_index, task=task
    )
    return raw


@pytest.fixture()
def data1(raw1):
    """Create merged free recall data."""
    data = fr.merge_free_recall(raw1, study_keys=['task', 'item_index'])
    return data


@pytest.fixture()
def data2(raw2):
    """Create merged free recall data."""
    data = fr.merge_free_recall(raw2, study_keys=['task', 'item_index'])
    return data


@pytest.fixture()
def stats_def():
    stats_def = analysis.Statistics()
    stats_def.set_stat(
        "spc", "psifr.fr:spc", ["input"], "recall", "group", conditions=["task"]
    )
    stats_def.set_stat(
        "lag_crp", "psifr.fr:lag_crp", ["lag"], "prob", "subject", conditions=["task"]
    )
    stats_def.set_stat(
        "lag_crp_compound", 
        "psifr.fr:lag_crp_compound", 
        ["previous", "current"], 
        "prob", 
        "subject", 
        conditions=["task"],
    )
    return stats_def


def test_eval_stats(data1, stats_def):
    """Test evaluation of statistics."""
    stats, results = stats_def.eval_stats(data1)
    spc = [0.5, 1, 1, 0.5, 0.5, 1]
    c1s1 = [np.nan, 0, np.nan, 1, np.nan]
    c1s2 = [np.nan, np.nan, np.nan, 1, 0]
    c2s1 = [1, 0, np.nan, np.nan, np.nan]
    c2s2 = [np.nan, 1, np.nan, 0, 1]
    dependent_expected = np.array(spc + c1s1 + c1s2 + c2s1 + c2s2)
    dependent = stats.loc[:25, "dependent"].to_numpy()
    np.testing.assert_array_equal(dependent, dependent_expected)


def test_compare_stats(data1, data2, stats_def):
    """Test comparison of stats for two datasets."""
    stats1, _ = stats_def.eval_stats(data1)
    stats2, _ = stats_def.eval_stats(data2)
    metric = stats_def.compare_stats(stats1, stats2)
    assert round(metric, 3) == 0.764



import os
from setuptools import setup, Extension
from Cython.Build import cythonize

ext_bases = ['src/cymr/operations', 'src/cymr/lba']
ext_list = []
for ext_base in ext_bases:
    ext_name = 'cymr.' + os.path.basename(ext_base)
    if os.path.exists(ext_base + '.pyx'):
        # for simplicity, always use .pyx if available
        ext = cythonize(ext_base + '.pyx')[0]
    elif os.path.exists(ext_base + '.c'):
        # if no .pyx and have .c, create an extension from that
        ext = Extension(ext_name, [ext_base + '.c'])
    else:
        raise IOError(f'No file for extension {ext_name} matching {ext_base}.*')
    ext_list.append(ext)

setup(ext_modules=ext_list)



"""Test operation of models of free recall."""

import os
import numpy as np
import pandas as pd
import pytest
from cymr import fit
from cymr import cmr
from cymr import analysis


@pytest.fixture()
def data():
    """Generate sample free recall data."""
    data = fit.sample_data('sample1')
    data['list'] = data['subject']
    return data


def test_prepare_lists(data):
    """Test splitting lists into study and recall list format."""
    study, recall = fit.prepare_lists(data)
    np.testing.assert_array_equal(study['input'][0], np.array([0, 1, 2]))
    np.testing.assert_array_equal(study['input'][1], np.array([0, 1, 2]))
    np.testing.assert_array_equal(study['item_index'][0], np.array([0, 1, 2]))
    np.testing.assert_array_equal(study['item_index'][1], np.array([3, 4, 5]))
    np.testing.assert_array_equal(recall['input'][0], np.array([1, 2]))
    np.testing.assert_array_equal(recall['input'][1], np.array([2, 0]))


def test_prepare_study(data):
    """Test splitting study lists."""
    study_data = data.loc[data['trial_type'] == 'study'].copy()
    study = fit.prepare_study(study_data)
    np.testing.assert_array_equal(study['position'][0], np.array([0, 1, 2]))
    np.testing.assert_array_equal(study['position'][1], np.array([0, 1, 2]))
    np.testing.assert_array_equal(study['item_index'][0], np.array([0, 1, 2]))
    np.testing.assert_array_equal(study['item_index'][1], np.array([3, 4, 5]))


def test_cmr(data):
    """Test CMR likelihood evaluation."""
    model = cmr.CMR()
    param = {
        'B_enc': 0.5,
        'B_rec': 0.8,
        'B_start': 0,
        'Afc': 0,
        'Dfc': 1,
        'Acf': 0,
        'Dcf': 1,
        'Lfc': 1,
        'Lcf': 1,
        'P1': 0,
        'P2': 1,
        'T': 10,
        'X1': 0.05,
        'X2': 1,
    }
    param_def, patterns = cmr.config_loc_cmr(6)
    stats = model.likelihood(data, param, param_def=param_def, patterns=patterns)
    np.testing.assert_allclose(stats['logl'].sum(), -5.936799964636842)
    assert stats['n'].sum() == 6


@pytest.fixture()
def param():
    """Generate a parameter definition with standard fixed values."""
    param = dict(
        B_rec=0.8,
        B_start=0,
        Afc=0,
        Dfc=1,
        Acf=0,
        Dcf=1,
        Lfc=1,
        Lcf=1,
        P1=0,
        P2=1,
        T=10,
        X1=0.05,
        X2=1,
    )
    return param


def test_cmr_fit(data, param):
    """Test fit of CMR parameters to sample data."""
    model = cmr.CMR()
    n_item = data['item_index'].nunique()
    param_def, patterns = cmr.config_loc_cmr(n_item)
    param_def.set_fixed(param)
    param_def.set_free(B_enc=(0, 1))
    results = model.fit_indiv(data, param_def, patterns=patterns, n_jobs=2)
    np.testing.assert_allclose(
        results['B_enc'].to_numpy(), np.array([0.72728744, 0.99883425]), atol=0.02
    )
    np.testing.assert_array_equal(results['n'].to_numpy(), [3, 3])


def test_cmr_fit_stat(data, param):
    """Test fit of CMR parameters to sample data."""
    model = cmr.CMR()
    n_item = data['item_index'].nunique()
    param_def, patterns = cmr.config_loc_cmr(n_item)
    param_def.set_fixed(param)
    param_def.set_free(B_enc=(0.2, 0.4))
    items = data.query('trial_type == "study"').sort_values('item_index')['item'].to_numpy()
    patterns['items'] = items

    stats_def = analysis.Statistics(error_stat='rmsd', weighting='point')
    stats_def.set_stat("spc", "psifr.fr:spc", ["input"], "recall", "group")

    results = model.fit_indiv(
        data, 
        param_def, 
        patterns=patterns, 
        stats_def=stats_def, 
        n_stats_rep=20, 
        n_jobs=2, 
        tol=0.1,
    )
    assert 'B_enc' in results
    assert 'rmsd' in results
    np.testing.assert_array_equal(results['n'].to_numpy(), [3, 3])


@pytest.fixture()
def patterns():
    """Generate patterns for use in CMR-D."""
    cat = np.array([[1, 0, 1, 1, 0, 1], [0, 1, 0, 0, 1, 0]]).T
    items = np.array(['absence', 'hollow', 'pupil', 'fountain', 'piano', 'pillow'])
    patterns = {
        'items': items,
        'vector': {'loc': np.eye(6), 'cat': cat},
        'similarity': {'loc': np.eye(6), 'cat': np.dot(cat, cat.T)},
    }
    return patterns


def test_pattern_io(patterns):
    temp = 'test_pattern.hdf5'
    cmr.save_patterns(
        temp,
        patterns['items'],
        loc=patterns['vector']['loc'],
        cat=patterns['vector']['cat'],
    )
    pat = cmr.load_patterns(temp)

    # vector representation
    expected = np.array([[1, 0], [0, 1], [1, 0], [1, 0], [0, 1], [1, 0]])
    np.testing.assert_allclose(pat['vector']['cat'], expected)

    # similarity matrix
    expected = np.array(
        [
            [1, 0, 1, 1, 0, 1],
            [0, 1, 0, 0, 1, 0],
            [1, 0, 1, 1, 0, 1],
            [1, 0, 1, 1, 0, 1],
            [0, 1, 0, 0, 1, 0],
            [1, 0, 1, 1, 0, 1],
        ]
    )
    np.testing.assert_allclose(pat['similarity']['cat'], expected)
    os.remove(temp)


def test_init_network(patterns):
    """Test initialization of a complex network."""
    param_def = cmr.CMRParameters()
    param_def.set_dependent(
        w_loc='wr_loc / sqrt(wr_loc**2 + wr_cat**2)',
        w_cat='wr_cat / sqrt(wr_loc**2 + wr_cat**2)',
        s_loc='sr_loc / (sr_loc + sr_cat)',
        s_cat='sr_cat / (sr_loc + sr_cat)',
    )
    param_def.set_sublayers(f=['task'], c=['loc', 'cat'])
    param_def.set_weights(
        'fc',
        {
            (('task', 'item'), ('loc', 'item')): 'w_loc * loc',
            (('task', 'item'), ('cat', 'item')): 'w_cat * cat',
        },
    )
    param_def.set_weights(
        'cf',
        {
            (('task', 'item'), ('loc', 'item')): 'w_loc * loc',
            (('task', 'item'), ('cat', 'item')): 'w_cat * cat',
        },
    )
    param_def.set_weights('ff', {('task', 'item'): 's_loc * loc + s_cat * cat'})
    item_index = np.arange(3)
    param = {'wr_loc': 1, 'wr_cat': np.sqrt(2), 'sr_loc': 1, 'sr_cat': 2}
    param = param_def.eval_dependent(param)
    net = cmr.init_network(param_def, patterns, param, item_index)

    expected = np.array(
        [
            [0.5774, 0.0000, 0.0000, 0, 0, 0, 0.0000, 0.8165, 0.0000, 0.0000],
            [0.0000, 0.5774, 0.0000, 0, 0, 0, 0.0000, 0.0000, 0.8165, 0.0000],
            [0.0000, 0.0000, 0.5774, 0, 0, 0, 0.0000, 0.8165, 0.0000, 0.0000],
            [0.0000, 0.0000, 0.0000, 0, 0, 0, 1.0000, 0.0000, 0.0000, 1.0000],
        ]
    )
    np.testing.assert_allclose(net.w_fc_pre, expected, atol=0.0001)

    expected = np.array(
        [
            [1.0000, 0.0000, 0.6667, 0.0000],
            [0.0000, 1.0000, 0.0000, 0.0000],
            [0.6667, 0.0000, 1.0000, 0.0000],
            [0.0000, 0.0000, 0.0000, 0.0000],
        ]
    )
    np.testing.assert_allclose(net.w_ff_pre, expected, atol=0.0001)


def test_init_network_distract():
    """Test initialization of a network with distraction units."""
    param_def = cmr.CMRParameters()
    param_def.set_options(distraction=True)
    param_def.set_sublayers(f=['task'], c=['loc'])
    param_def.set_weights('fc', {(('task', 'item'), ('loc', 'item')): 'loc'})
    param_def.set_weights('cf', {(('task', 'item'), ('loc', 'item')): 'loc'})
    item_index = np.arange(3)
    param = {}
    patterns = {'vector': {'loc': np.eye(3)}}
    net = cmr.init_network(param_def, patterns, param, item_index)
    np.testing.assert_array_equal(net.w_fc_pre, np.eye(8))


@pytest.fixture()
def param_def_dist(param):
    """Generate parameter definitions for a simple CMR-D network."""
    param_def = cmr.CMRParameters()
    param_def.set_fixed(param)
    param_def.set_sublayers(f=['task'], c=['task'])
    weights = {(('task', 'item'), ('task', 'item')): 'loc'}
    param_def.set_weights('fc', weights)
    param_def.set_weights('cf', weights)
    return param_def


@pytest.fixture()
def param_dist():
    """Generate standard parameters for testing CMR-D."""
    param = {
        'B_enc': 0.5,
        'B_start': 0,
        'B_rec': 0.8,
        'Lfc': 1,
        'Lcf': 1,
        'P1': 0,
        'P2': 1,
        'T': 10,
        'X1': 0.05,
        'X2': 1,
    }
    return param


def test_dist_cmr(data, patterns, param_def_dist, param_dist):
    """Test localist CMR using the CMR-D implementation."""
    model = cmr.CMR()
    stats = model.likelihood(
        data, param_dist, None, param_def_dist, patterns=patterns
    )
    np.testing.assert_allclose(stats['logl'].sum(), -5.936799964636842)


def test_dist_cmr_distract(data, patterns, param_def_dist, param_dist):
    """Test CMR-D with distraction."""
    model = cmr.CMR()
    param_def = param_def_dist.copy()
    param_def.set_options(distraction=True)
    param = param_dist.copy()
    param['B_distract'] = 0
    param['B_retention'] = 0
    stats = model.likelihood(data, param, None, param_def, patterns=patterns)
    np.testing.assert_allclose(stats['logl'].sum(), -5.936799964636842)

    param['B_distract'] = 0.2
    param['B_retention'] = 0.2
    stats = model.likelihood(data, param, None, param_def, patterns=patterns)
    np.testing.assert_allclose(stats['logl'].sum(), -5.933655795592753)


def test_dist_cmr_fit(data, patterns, param_def_dist):
    """Test fitted parameter values for CMR-D."""
    param_def = param_def_dist.copy()
    model = cmr.CMR()
    param_def.set_fixed(w_loc=1)
    param_def.set_free(B_enc=(0, 1))
    results = model.fit_indiv(data, param_def, patterns=patterns, n_jobs=2)
    np.testing.assert_allclose(
        results['B_enc'].to_numpy(), np.array([0.72728744, 0.99883425]), atol=0.02
    )


def test_dist_cmr_generate(data, patterns, param_def_dist, param_dist):
    """Test that CMR-D generation runs."""
    model = cmr.CMR()
    sim = model.generate(data, param_dist, None, param_def_dist, patterns=patterns)
    assert isinstance(sim, pd.DataFrame)


def test_dist_cmr_record(data, patterns, param_def_dist, param_dist):
    model = cmr.CMR()
    states = model.record(data, param_dist, None, param_def_dist, patterns=patterns)

    # test the series of context states
    np.testing.assert_allclose(
        states[0].c, np.array([0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8660254])
    )
    np.testing.assert_allclose(
        states[1].c, np.array([0.4330127, 0.5, 0.0, 0.0, 0.0, 0.0, 0.75])
    )
    np.testing.assert_allclose(
        states[2].c, np.array([0.375, 0.4330127, 0.5, 0.0, 0.0, 0.0, 0.64951905])
    )
    np.testing.assert_allclose(
        states[3].c, np.array([0.375, 0.4330127, 0.5, 0.0, 0.0, 0.0, 0.64951905])
    )
    np.testing.assert_allclose(
        states[4].c,
        np.array([0.29319805, 0.80043616, 0.12426407, 0.0, 0.0, 0.0, 0.50783392]),
    )

    # test various network components
    np.testing.assert_allclose(
        states[5].c, np.array([0.0, 0.0, 0.0, 0.5, 0.0, 0.0, 0.8660254])
    )
    np.testing.assert_allclose(
        states[5].c_in, np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0])
    )
    np.testing.assert_allclose(states[5].f, np.array([1.0, 0.0, 0.0, 0.0]))
    np.testing.assert_allclose(states[5].f_in, np.array([0.0, 0.0, 0.0, 0.0]))


def test_dist_cmr_distract_record(data, patterns, param_def_dist, param_dist):
    model = cmr.CMR()
    param_def = param_def_dist.copy()
    param_def.set_options(distraction=True)
    param = param_dist.copy()
    param['B_distract'] = 0.2
    param['B_retention'] = 0.4
    states = model.record(data, param, None, param_def, patterns=patterns)

    np.testing.assert_allclose(
        states[0].c, np.array([0.5, 0, 0, 0, 0, 0, 0.84852814, 0.17320508, 0, 0, 0])
    )
    np.testing.assert_allclose(
        states[1].c, 
        np.array([0.42426407, 0.5, 0, 0, 0, 0, 0.72, 0.14696938, 0.17320508, 0, 0])
    )
    np.testing.assert_allclose(
        states[2].c, 
        np.array([0.36, 0.42426407, 0.5, 0, 0, 0, 0.61094026, 0.12470766, 0.14696938, 0.17320508, 0])
    )


def test_dist_cmr_record_trim(data, patterns, param_def_dist, param_dist):
    model = cmr.CMR()
    states = model.record(
        data, param_dist, None, param_def_dist, patterns=patterns, remove_blank=True
    )
    np.testing.assert_allclose(states[0].c, np.array([0.5, 0.0, 0.0, 0.8660254]))


def test_dist_cmr_record_include(data, patterns, param_def_dist, param_dist):
    model = cmr.CMR()
    states = model.record(
        data, param_dist, None, param_def_dist, patterns=patterns, include=['c', 'c_in']
    )
    assert hasattr(states[0], 'c') and hasattr(states[0], 'c_in')
    assert not hasattr(states[0], 'f')


def test_dynamic_cmr(data, patterns, param_def_dist, param_dist):
    """Test evaluation of a dynamic study parameter."""
    param = param_dist.copy()
    param_def = param_def_dist.copy()
    param['slope_distract'] = 0.2
    param_def.set_dynamic(
        'study', 
        'item',
        B_distract='distract * slope_distract', 
        B_retention='retention * slope_distract',
    )
    param_def.set_options(distraction=True)
    
    model = cmr.CMR()
    stats = model.likelihood(
        data, 
        param, 
        None, 
        param_def, 
        patterns=patterns, 
        study_keys=['distract', 'retention'],
    )
    np.testing.assert_allclose(stats['logl'].sum(), -5.945280418642251)


def test_dynamic_cmr_recall(data, patterns, param_def_dist, param_dist):
    """Test evaluation of a dynamic recall parameter."""
    param = param_dist.copy()
    param_def = param_def_dist.copy()
    param['B_op'] = 0.2
    param_def.set_dynamic('recall', 'item', B_rec='op * B_op')
    model = cmr.CMR()
    stats = model.likelihood(
        data, param, None, param_def, patterns=patterns, recall_keys=['op']
    )
    np.testing.assert_allclose(stats['logl'].sum(), -5.919470385031945)


@pytest.fixture()
def param_def_sublayer():
    """Generate parameter definitions for multiple context sublayers."""
    param_def = cmr.CMRParameters()
    param_def.set_sublayers(f=['task'], c=['loc', 'cat'])
    param_def.set_sublayer_param('c', 'loc', {'B_enc': 'B_enc_loc'})
    param_def.set_sublayer_param('c', 'cat', {'B_enc': 'B_enc_cat'})
    weights = {
        (('task', 'item'), ('loc', 'item')): 'loc',
        (('task', 'item'), ('cat', 'item')): 'cat',
    }
    param_def.set_weights('fc', weights)
    param_def.set_weights('cf', weights)
    return param_def


def test_sublayer_study(data, patterns, param_def_sublayer, param_dist):
    """Test steps involved in working with context sublayers."""
    # test expanded list parameters
    param = param_dist.copy()
    param['B_enc_loc'] = 0.5
    param['B_enc_cat'] = 0.8
    n_item = 3
    n_sub = 2
    list_param = cmr.prepare_list_param(n_item, n_sub, param, param_def_sublayer)
    np.testing.assert_array_equal(list_param['Lfc'], np.array([[1, 1], [1, 1], [1, 1]]))
    np.testing.assert_array_equal(list_param['Lcf'], np.array([[1, 1], [1, 1], [1, 1]]))

    # test sublayer-specific parameters
    assert param_def_sublayer.sublayers['c'] == ['loc', 'cat']
    list_param = param_def_sublayer.eval_sublayer_param('c', list_param, n_item)
    np.testing.assert_array_equal(
        list_param['B_enc'], np.array([[0.5, 0.8], [0.5, 0.8], [0.5, 0.8]])
    )

    # prepare lists for simulation
    study, recall = fit.prepare_lists(
        data, study_keys=['input', 'item_index'], recall_keys=['input'], clean=True
    )
    item_index = np.arange(len(patterns['items']))
    i = 0
    item_pool, item_study, item_recall, item_distract = cmr.get_list_items(
        item_index, study, recall, i, 'list'
    )

    # study the first list
    net = cmr.study_list(
        param_def_sublayer,
        list_param,
        item_pool,
        item_study,
        patterns,
        item_distract,
    )
    expected = np.array(
        [
            [0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8660, 0.8, 0.0, 0.6],
            [0.4330, 0.5, 0.0, 0.0, 0.0, 0.0, 0.75, 0.48, 0.8, 0.36],
            [0.375, 0.4330, 0.5, 0.0, 0.0, 0.0, 0.6495, 0.9576, 0.2627, 0.1182],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )
    np.testing.assert_allclose(net.w_fc_exp, expected, atol=0.0001)


def test_sublayer_cmr(data, patterns, param_def_sublayer, param_dist):
    """Test evaluation of CMR-D with multiple context sublayers."""
    param = param_dist.copy()
    param['B_enc_loc'] = 0.5
    param['B_enc_cat'] = 0.8
    model = cmr.CMR()
    stats = model.likelihood(data, param, None, param_def_sublayer, patterns=patterns)
    np.testing.assert_allclose(stats['logl'].sum(), -5.8694368046085215)


def test_sublayer_generate(data, patterns, param_def_sublayer, param_dist):
    """Test using CMR-D with sublayers to generate data."""
    param = param_dist.copy()
    param['B_enc_loc'] = 0.5
    param['B_enc_cat'] = 0.8
    model = cmr.CMR()

    # recall from list items only
    param_def_sublayer.set_options(scope='list')
    sim = model.generate(data, param, None, param_def_sublayer, patterns=patterns)
    assert isinstance(sim, pd.DataFrame)

    # recall from the pool
    param_def_sublayer.set_options(scope='pool')
    sim = model.generate(data, param, None, param_def_sublayer, patterns=patterns)
    assert isinstance(sim, pd.DataFrame)


def test_match_generate(data, patterns, param_def_dist, param_dist):
    """Test generating recalls using context match screening."""
    param = param_dist.copy()
    param['A1'] = -3
    param['A2'] = 1
    model = cmr.CMR()
    param_def_dist.set_options(scope='pool', filter_recalls=True)
    sim = model.generate(data, param, None, param_def_dist, patterns=patterns)
    assert isinstance(sim, pd.DataFrame)



import numpy as np
import pandas as pd

from cymr import operations
from cymr import fit
from cymr import network
from cymr import cmr
from cymr.cmr import CMR
from cymr import lba


def add_recalls(study, recalls_list, recalls_times):
    """Doc"""
    lists = study['list'].unique()
    subjects = study['subject'].unique()
    if len(subjects) > 1:
        raise ValueError('Unpacking multiple subjects not supported.')
    subject = subjects[0]

    # initialize recall trials DataFrame
    n_recall = np.sum([len(r) for r in recalls_list])
    recall = pd.DataFrame(
        {
            'subject': subject,
            'list': np.zeros(n_recall, dtype=int),
            'trial_type': 'recall',
            'position': np.zeros(n_recall, dtype=int),
            'item': '',
            'rt': np.zeros(n_recall, dtype=float),
        }
    )

    # set basic information (list, item, position)
    n = 0
    for i, seq in enumerate(recalls_list):
        pool = study.loc[study['list'] == lists[i], 'item'].to_numpy()
        for j, pos in enumerate(seq):
            recall.loc[n, 'list'] = lists[i]
            recall.loc[n, 'item'] = pool[pos]
            recall.loc[n, 'position'] = j + 1
            recall.loc[n, 'rt'] = recalls_times[i][j]
            n += 1
    data = pd.concat((study, recall), axis=0, ignore_index=True)
    data = data.sort_values(['list', 'trial_type'], ascending=[True, False])
    return data


class CMRLBA(CMR):
    """CMR Distributed with Linear Ballistic Accumulators."""

    def prepare_sim(self, data, param):
        # 'dynamic' field on param
        # tells you where the dynamic param values are stored on the data struct
        study_key_names = ['input', 'item_index']
        recall_key_names = ['input', 'rt']
        if 'dynamic' in param:
            if 'study' in param['dynamic']:
                for pname in param['dynamic']['study'].keys():
                    datacol = param['dynamic']['study'][pname]
                    study_key_names = study_key_names + datacol
            if 'recall' in param['dynamic']:
                for pname in param['dynamic']['recall'].keys():
                    datacol = param['dynamic']['recall'][pname]
                    recall_key_names = recall_key_names + datacol
        study, recall = fit.prepare_lists(
            data, study_keys=study_key_names, recall_keys=recall_key_names, clean=True
        )
        return study, recall

    def generate_subject(
        self, study_data, param, patterns=None, weights=None, **kwargs
    ):
        study = fit.prepare_study(study_data, study_keys=['position', 'item_index'])
        n_item = len(study['position'][0])
        n_list = len(study['position'])
        list_param = cmr.prepare_list_param(n_item, param)

        weights_param = network.unpack_weights(weights, param)
        scaled = network.prepare_patterns(patterns, weights_param)
        recalls = []
        rts = []
        recall_time_limit = param['recall_time_limit']
        for i in range(n_list):
            net = cmr.init_dist_cmr(study['item_index'][i], scaled, param)
            net.study(
                'item',
                study['position'][i],
                param['B_enc'],
                list_param['Lfc'],
                list_param['Lcf'],
            )
            net.integrate('start', 0, param['B_start'])
            recall, rt = net.generate_recall_lba(
                'item',
                recall_time_limit,
                param['B_rec'],
                param['A'],
                param['b'],
                param['s'],
                param['tau'],
            )
            recalls.append(recall)
            rts.append(rt)
        data = add_recalls(study_data, recalls, rts)
        return data

    def likelihood_subject(self, study, recall, param, patterns=None, weights=None):
        n_item = len(study['input'][0])
        n_list = len(study['input'])
        list_param = cmr.prepare_list_param(n_item, param)

        weights_param = network.unpack_weights(weights, param)
        scaled = network.prepare_patterns(patterns, weights_param)
        logl = 0
        n = 0
        for i in range(n_list):
            net = cmr.init_dist_cmr(study['item_index'][i], scaled, param)
            sparam = param.copy()
            if 'dynamic' in param:
                if 'study' in param['dynamic']:
                    sparam = cmr.update_dynamic_parameters(study, param, 'study', i)
            net.study(
                'item',
                study['input'][i],
                sparam['B_enc'],
                list_param['Lfc'],
                list_param['Lcf'],
            )
            net.integrate('start', 0, param['B_start'])
            rparam = param.copy()
            if 'dynamic' in param:
                if 'recall' in param['dynamic']:
                    rparam = cmr.update_dynamic_parameters(recall, param, 'recall', i)
            p = self.p_recall_lba(
                net,
                'item',
                recall['input'][i],
                recall['rt'][i],
                param['recall_time_limit'],
                rparam['B_rec'],
                param['T'],
                param['A'],
                param['b'],
                param['s'],
                param['tau'],
            )
            if np.any(np.isnan(p)) or np.any((p <= 0) | (p >= 1)):
                logl = -10e6
                break
            logl += np.sum(np.log(p))
            n += p.size
        return logl, n

    def p_recall_lba(
        self, net, segment, recalls, rts, recall_time_limit, B, T, A, b, s, tau
    ):
        """Recall data struct must contain rt information for recall events."""
        amin = 0.000001
        # weights to use for recall (assume fixed during recall)
        rec_ind = net.f_ind[segment]
        n_item = net.n_f_segment[segment]
        w_cf = net.w_cf_exp[rec_ind, :] + net.w_cf_pre[rec_ind, :]

        exclude = np.zeros(n_item, dtype=np.dtype('i'))
        # exclude = np.zeros(w_cf.shape[0], dtype=bool)
        p = np.zeros(len(recalls) + 1)
        for output, recall in enumerate(recalls):
            # project the current state of context; assume nonzero support
            # support = np.dot(w_cf, net.c)
            # support[support < amin] = amin

            # calculate item support (modifies f_in)
            operations.cue_item(
                rec_ind.start,
                n_item,
                net.w_cf_pre,
                net.w_cf_exp,
                net.w_ff_pre,
                net.w_ff_exp,
                net.f_in,
                net.c,
                exclude,
                np.asarray(recalls, dtype=np.dtype('i')),
                output,
            )
            support = net.f_in[rec_ind]

            # TODO: allow some kind of non-linear scaling of support
            # scale based on choice parameter, set recalled items to zero
            # strength = np.exp((2 * support) / T)
            # strength[exclude] = 0
            strength = support
            strength[exclude] = 0

            # probability of this recall at this rt, conditional on not stopping
            if output > 0:
                resp_duration = rts[output] - rts[output - 1]
            else:
                resp_duration = rts[output]
            p[output] = lba.response_pdf(resp_duration, recall, A, b, strength, s)

            # remove recalled item from competition
            exclude[recall] = 1

            # update context
            net.present(segment, recall, B)

        # probability of the stopping event
        # is the probability that no racers cross threshold before time is up
        time_left = recall_time_limit - rts[-1]

        p[len(recalls)] = lba.response_pdf(time_left, len(strength), A, b, strength, s)
        return p



"""Represent interactions between context and item layers."""

import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from cymr import operations


def expand_param(param, size):
    """
    Expand a scalar parameter to array format.

    Parameters
    ----------
    param : float or numpy.ndarray
        Parameter to expand.

    size : iterable or numpy.ndarray
        Size of the expanded parameter.

    Returns
    -------
    param : numpy.ndarray
        Expanded parameter.
    """
    size = np.asarray(size)
    if not isinstance(param, np.ndarray):
        # expand scalar to full array
        param = np.tile(param, size).astype(float)
    elif size.shape and param.ndim < len(size):
        # expand array to have to correct number of dimensions
        param = np.expand_dims(param, axis=1)
        if param.shape != tuple(size):
            # expand singleton dimensions as needed
            rep = np.ones(size.shape, dtype=int)
            for i, n in enumerate(param.shape):
                if n == 1:
                    rep[i] = size[i]
                elif n != size[i] and n < size[i]:
                    raise ValueError('Cannot expand parameter.')
            param = np.tile(param, rep).astype(float)
    return param


def prepare_study_param(n_item, n_sub, B, Lfc, Lcf, distract_B=None, retention_B=None):
    """Prepare parameters for simulating a study phase."""
    B = expand_param(B, (n_item, n_sub))
    Lfc = expand_param(Lfc, (n_item, n_sub))
    Lcf = expand_param(Lcf, (n_item, n_sub))
    param = {'B': B, 'Lfc': Lfc, 'Lcf': Lcf, 'distract_B': None}
    if distract_B is not None:
        distract_B = expand_param(distract_B, (n_item, n_sub))
        if isinstance(retention_B, np.ndarray) and retention_B.shape[0] > 1:
            # if there is a value for each item, use the last one
            retention_B = retention_B[[-1]]
        retention_B = expand_param(retention_B, (1, n_sub))
        param['distract_B'] = np.concat([distract_B, retention_B], 0)
    return param


def prepare_recall_param(n_item, n_sub, B, T, amin):
    """Prepare parameters for simulating a recall phase."""
    B = expand_param(B, (n_item, n_sub))
    if T < amin:
        T = amin
    param = {'B': B, 'T': T}
    return param


def sample_response_lba(A, b, v, s, tau):
    """Sample a response and response time."""
    while True:
        k = stats.uniform.rvs(0, scale=A)
        d = stats.norm.rvs(loc=v, scale=s)
        t = tau + (b - k) / d
        if np.any(d > 0):
            break
    t[d <= 0] = np.nan
    response = np.nanargmin(t)
    rt = np.nanmin(t)
    return int(response), rt


def init_plot(**kwargs):
    fig = plt.figure(constrained_layout=True, **kwargs)
    gs = GridSpec(10, 5, figure=fig)
    ax = {
        'c': fig.add_subplot(gs[0, 1:4]),
        'c_in': fig.add_subplot(gs[1, 1:4]),
        'f_in': fig.add_subplot(gs[8, 1:4]),
        'f': fig.add_subplot(gs[9, 1:4]),
        'w_fc_pre': fig.add_subplot(gs[2:8, 0]),
        'w_fc_exp': fig.add_subplot(gs[2:8, 1]),
        'w_ff_pre': fig.add_subplot(gs[2:8, 2]),
        'w_cf_exp': fig.add_subplot(gs[2:8, 3]),
        'w_cf_pre': fig.add_subplot(gs[2:8, 4]),
    }
    return fig, ax


class LayerIndex(object):
    """
    Representation of one layer of a network.

    Parameters
    ----------
    layer_segments : dict of (str: dict of (str: int))
        Size of each named segment for each sublayer.

    Attributes
    ----------
    size : int
        Number of units in the layer.

    size_sublayer : dict of (str: int)
        Number of units in each sublayer.

    size_segment : dict of (str: dict of (str: int))
        Size of each segment in each sublayer.

    sublayer : dict of (str: numpy.ndarray)
        Start and stop units of each sublayer.

    segment : dict of (str: dict of (str: numpy.ndarray)
        Start and stop units of each segment.
    """

    def __init__(self, layer_segments):
        self.size = 0
        self.size_sublayer = {}
        self.size_segment = layer_segments.copy()
        self.sublayer = {}
        self.segment = {}
        for sub, segs in layer_segments.items():
            self.segment[sub] = {}
            self.size_sublayer[sub] = 0
            start = self.size
            for seg, s in segs.items():
                self.segment[sub][seg] = np.array(
                    [self.size, self.size + s], dtype=np.dtype('i')
                )
                self.size += s
                self.size_sublayer[sub] += s
            self.sublayer[sub] = np.array(
                [start, start + self.size_sublayer[sub]], dtype=np.dtype('i')
            )

    def __repr__(self):
        s = ''
        for sublayer, segments in self.size_segment.items():
            size_sublayer = self.size_sublayer[sublayer]
            s += f'{sublayer}: {size_sublayer} units\n'
            for segment, size_segment in segments.items():
                s += f'    {segment}: {size_segment} units\n'
        return s

    def copy(self):
        return self.__init__(self.size_segment)

    def get_sublayer(self, sublayer):
        return self.sublayer[sublayer]

    def get_segment(self, sublayer, segment):
        return self.segment[sublayer][segment]

    def get_slice(self, sublayer, segment):
        ind = self.get_segment(sublayer, segment)
        return slice(ind[0], ind[1])

    def get_unit(self, sublayer, segment, index):
        return self.segment[sublayer][segment][0] + index


class Network(object):
    """
    Representation of interacting item and context layers.

    Parameters
    ----------
    f_segment : dict of str: (dict of str: int)
        For each item sublayer, the number of units in each segment.

    c_segment : dict of str: (dict of str: int)
        For each context sublayer, the number of units in each segment.

    Attributes
    ----------
    f_segment : dict of str: (dict of str: int)
        Number of item units for each segment.

    c_segment : dict of str: (dict of str: int)
        Number of context units for each segment.

    f_ind : cymr.network.LayerIndex
        Index of units in the item layer.

    c_ind : cymr.network.LayerIndex
        Index of units in the context layer.

    n_f : int
        Total number of item units.

    n_c : int
        Total number of context units.

    f : numpy.array
        Item layer vector.

    c : numpy.array
        Context layer vector.

    c_in : numpy.array
        Current input to context.

    w_fc_pre : numpy.array
        Pre-experimental weights connecting f to c.

    w_fc_exp : numpy.array
        Weights learned during the experiment connecting f to c.

    w_cf_pre : numpy.array
        Pre-experimental weights connecting c to f.

    w_cf_exp : numpy.array
        Weights learned during the experiment connecting c to f.

    w_ff_pre : numpy.array
        Pre-experimental weights connecting f to f.

    w_ff_exp : numpy.array
        Weights learned during the experiment connecting f to f.
    """

    def __init__(self, f_segment, c_segment):
        self.f_segment = f_segment.copy()
        self.c_segment = c_segment.copy()
        self.f_sublayers = list(self.f_segment.keys())
        self.c_sublayers = list(self.c_segment.keys())
        self.f_ind = LayerIndex(self.f_segment)
        self.c_ind = LayerIndex(self.c_segment)
        n_f = self.f_ind.size
        n_c = self.c_ind.size
        self.n_f = n_f
        self.n_c = n_c
        self.f = np.zeros(n_f)
        self.f_in = np.zeros(n_f)
        self.match = np.zeros(n_f)
        self.c = np.zeros(n_c)
        self.c_in = np.zeros(n_c)
        self.w_fc_pre = np.zeros((n_f, n_c))
        self.w_fc_exp = np.zeros((n_f, n_c))
        self.w_cf_pre = np.zeros((n_f, n_c))
        self.w_cf_exp = np.zeros((n_f, n_c))
        self.w_ff_pre = np.zeros((n_f, n_f))
        self.w_ff_exp = np.zeros((n_f, n_f))

    def __repr__(self):
        s = f'f:\n{self.f_ind}\nc:\n{self.c_ind}'
        return s

    def reset(self):
        """Reset network weights and activations to zero."""
        self.f[:] = 0
        self.f_in[:] = 0
        self.match[:] = 0
        self.c[:] = 0
        self.c_in[:] = 0
        self.w_fc_exp[:] = 0
        self.w_fc_pre[:] = 0
        self.w_cf_exp[:] = 0
        self.w_cf_pre[:] = 0
        self.w_ff_exp[:] = 0
        self.w_ff_pre[:] = 0

    def copy(self, include=None, exclude=None):
        """
        Copy the network to a new network object.

        Parameters
        ----------
        include : list of str, optional
            List of fields to include. Default is a standard list of
            fields.

        exclude : list of str, optional
            List of fields to exclude. Overrides values for include.

        Returns
        -------
        net : cymr.Network
            Network with the same segments, weights, and activations.
        """
        # all attributes that contain an array
        fields = [
            'f',
            'f_in',
            'match',
            'c',
            'c_in',
            'w_fc_exp',
            'w_fc_pre',
            'w_cf_exp',
            'w_cf_pre',
            'w_ff_exp',
            'w_ff_pre',
        ]

        # set the fields to include in the copy (default is all)
        if include is None:
            include = fields
        if exclude is not None:
            include = [i for i in include if i not in exclude]

        # copy included attributes
        net = Network(self.f_segment, self.c_segment)
        for field in fields:
            if field in include:
                setattr(net, field, getattr(self, field).copy())
            else:
                delattr(net, field)
        return net

    def get_sublayer(self, layer, sublayer):
        """
        Get indices for a sublayer.

        Parameters
        ----------
        layer : str
            Layer to access.

        sublayer : str
            Sublayer to access.

        Returns
        -------
        numpy.ndarray
            Start and stop indices for the sublayer.
        """
        if layer == 'f':
            ind = self.f_ind.get_sublayer(sublayer)
        elif layer == 'c':
            ind = self.c_ind.get_sublayer(sublayer)
        else:
            raise ValueError(f'Invalid layer: {layer}')
        return ind

    def get_sublayers(self, layer, sublayers):
        """
        Get an array of indices for multiple sublayers.

        Parameters
        ----------
        layer : str
            Layer to access.

        sublayers : list of str
            Sublayers to access.

        Returns
        -------
        numpy.array
            [sublayers x 2] array of start and stop indices for each
            sublayer.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        ind_list = []
        for sublayer in sublayers:
            ind_list.append(self.get_sublayer(layer, sublayer))
        return np.array(ind_list, dtype=np.dtype('i'))

    def get_segment(self, layer, sublayer, segment):
        """
        Get indices for a segment.

        Parameters
        ----------
        layer : {'f', 'c'}
            Layer to access.

        sublayer : str
            Sublayer to access.

        segment : str
            Segment to access.

        Returns
        -------
        ind : numpy.array
            Segment indices.
        """
        if layer == 'f':
            ind = self.f_ind.get_segment(sublayer, segment)
        elif layer == 'c':
            ind = self.c_ind.get_segment(sublayer, segment)
        else:
            raise ValueError(f'Invalid layer: {layer}')
        return ind

    def get_unit(self, layer, sublayer, segment, unit):
        """
        Get indices for a unit.

        Parameters
        ----------
        layer : {'f', 'c'}
            Layer to access.

        sublayer : str
            Sublayer to access.

        segment : str
            Segment to access.

        unit : int
            Index of the unit within the segment.

        Returns
        -------
        ind : int
            Absolute index of the unit within the layer.
        """
        if layer == 'f':
            ind = self.f_ind.get_unit(sublayer, segment, unit)
        elif layer == 'c':
            ind = self.c_ind.get_unit(sublayer, segment, unit)
        else:
            raise ValueError(f'Invalid layer: {layer}')
        return ind

    def get_slice(self, layer, sublayer, segment):
        """
        Get a slice for a segment.

        Parameters
        ----------
        layer : {'f', 'c'}
            Layer to access.

        sublayer : str
            Sublayer to access.

        segment : str
            Segment to access.

        Returns
        -------
        slice
            Slice for indexing the segment.
        """
        if layer == 'f':
            index = self.f_ind
        elif layer == 'c':
            index = self.c_ind
        else:
            raise ValueError(f'Invalid layer: {layer}')
        return index.get_slice(sublayer, segment)

    def get_region(self, f_segment, c_segment):
        """
        Return slices for a region.

        Parameters
        ----------
        f_segment : tuple of (str, str)
            Sublayer and segment of the item layer.

        c_segment : tuple of (str, str)
            Sublayer and segment of the context layer.

        Returns
        -------
        f_ind : slice
            Span of the region in the item dimension.

        c_ind : slice
            Span of the region in the context dimension.
        """
        f_ind = self.f_ind.get_slice(*f_segment)
        c_ind = self.c_ind.get_slice(*c_segment)
        return f_ind, c_ind

    def add_pre_weights(
        self, connect, f_segment, c_segment, weights, slope=1, intercept=0
    ):
        """
        Add pre-experimental weights to a network.

        Parameters
        ----------
        connect : {'fc', 'cf'}
            Connections to add weights to.

        f_segment : tuple of str, str
            Sublayer and segment to use for the item layer.

        c_segment : tuple of str, str
            Sublayer and segment to use for the context layer.

        weights : numpy.array
            Items x context array of weights.

        slope : double, optional
            Slope to multiply weights by before adding.

        intercept : double, optional
            Intercept to add to weights.
        """
        scaled = intercept + slope * weights
        if connect == 'ff':
            f_slice = slice(*tuple(self.f_ind.get_segment(*f_segment)))
            c_slice = None
        else:
            f_slice, c_slice = self.get_region(f_segment, c_segment)
        if connect == 'fc':
            self.w_fc_pre[f_slice, c_slice] = scaled
        elif connect == 'cf':
            self.w_cf_pre[f_slice, c_slice] = scaled
        elif connect == 'ff':
            self.w_ff_pre[f_slice, f_slice] = scaled
        else:
            raise ValueError(f'Invalid connection type: {connect}')

    def update(self, item, sublayers):
        """
        Update context completely with input from the item layer.

        Rather than integrating input into context, this replaces the
        current state of context with the input.

        Parameters
        ----------
        item : tuple of str, str, int
            Sublayer, segment, and unit of the item to present.

        sublayers : str of list of str
            Sublayer(s) of context to update.
        """
        f_ind = self.get_unit('f', *item)
        c_ind = self.get_sublayers('c', sublayers)
        B = np.ones(len(sublayers))
        operations.integrate(
            self.w_fc_exp, self.w_fc_pre, self.c, self.c_in, self.f, f_ind, c_ind, B
        )

    def integrate(self, item, sublayers, B):
        """
        Integrate input from the item layer into context.

        Parameters
        ----------
        item : tuple of str, str, int
            Sublayer, segment, and unit of the item to present.

        sublayers : str or list of str
            Sublayer(s) of context to update.

        B : float or numpy.array
            Integration scaling factor for each sublayer; higher values
            update context to more strongly reflect the input.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        B = expand_param(B, np.array(len(sublayers)))
        f_ind = self.get_unit('f', *item)
        c_ind = self.get_sublayers('c', sublayers)
        operations.integrate(
            self.w_fc_exp, self.w_fc_pre, self.c, self.c_in, self.f, f_ind, c_ind, B
        )

    def present(self, item, sublayers, B, Lfc=0, Lcf=0):
        """
        Present an item and learn context-item associations.

        Parameters
        ----------
        item : tuple of str, str, int
            Sublayer, segment, and unit of the item to present.

        sublayers : str or list of str
            Sublayer(s) of context to update.

        B : float or numpy.ndarray
            Integration scaling factors; higher values update context
            to more strongly reflect the input.

        Lfc : float or numpy.ndarray, optional
            Learning rates for item to context associations.

        Lcf : float or numpy.ndarray, optional
            Learning rates for context to item associations.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        n_sub = np.array(len(sublayers))
        B = expand_param(B, n_sub)
        Lfc = expand_param(Lfc, n_sub)
        Lcf = expand_param(Lcf, n_sub)
        f_ind = self.get_unit('f', *item)
        c_ind = self.get_sublayers('c', sublayers)
        operations.present(
            self.w_fc_exp,
            self.w_fc_pre,
            self.w_cf_exp,
            self.c,
            self.c_in,
            self.f,
            f_ind,
            c_ind,
            B,
            Lfc,
            Lcf,
        )

    def learn(self, connect, item, sublayers, L):
        """
        Learn an association between the item and context layers.

        Parameters
        ----------
        connect : {'fc', 'cf'}
            Connection matrix to update.

        item : tuple of str, str, int
            Sublayer, segment, and unit of the item to present.

        sublayers : str
            Sublayers of context to update.

        L : double
            Learning rate.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        L = expand_param(L, len(sublayers))
        f_ind = self.get_unit('f', *item)
        c_ind = self.get_sublayers('c', sublayers)
        if connect == 'fc':
            mat = self.w_fc_exp
        elif connect == 'cf':
            mat = self.w_cf_exp
        else:
            raise ValueError(f'Invalid connection: {connect}')
        for i, s_ind in enumerate(c_ind):
            s_slice = slice(*tuple(s_ind))
            mat[f_ind, s_slice] += self.c[s_slice] * L[i]

    def study(self, segment, item_list, sublayers, B, Lfc, Lcf):
        """
        Study a list of items.

        Parameters
        ----------
        segment : tuple of str, str
            Sublayer and segment of items to present.

        item_list : numpy.array
            Item indices relative to the segment.

        sublayers : str or list of str
            Sublayers of context to update.

        B : float or numpy.array
            Context updating rate. If an array, specifies a rate for
            each individual study trial.

        Lfc : float or numpy.array
            Learning rate for item to context associations. If an
            array, specifies a learning rate for each individual trial.

        Lcf : float or numpy.array
            Learning rate for context to item associations.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        param = prepare_study_param(item_list.shape[0], len(sublayers), B, Lfc, Lcf)

        # get unit indices
        f_ind = self.get_segment('f', *segment)
        item_ind = (f_ind[0] + item_list).astype(np.dtype('i'))
        c_ind = self.get_sublayers('c', sublayers)

        operations.study(
            self.w_fc_exp,
            self.w_fc_pre,
            self.w_cf_exp,
            self.c,
            self.c_in,
            self.f,
            item_ind,
            c_ind,
            param['B'],
            param['Lfc'],
            param['Lcf'],
        )

    def study_distract(
        self,
        segment,
        item_list,
        sublayers,
        B,
        Lfc,
        Lcf,
        distract_segment,
        distract_list,
        distract_B,
        retention_B,
    ):
        """
        Study a list of items.

        Parameters
        ----------
        segment : tuple of str, str
            Sublayer and segment representing the items to be presented.

        item_list : numpy.array
            Item indices relative to the segment.

        sublayers : str or list of str
            Sublayers to update during study.

        B : float or numpy.array
            Context updating rate. If an array, specifies a rate for
            each individual study trial.

        Lfc : float or numpy.array
            Learning rate for item to context associations. If an
            array, specifies a learning rate for each individual trial.

        Lcf : float or numpy.array
            Learning rate for context to item associations.

        distract_segment : str, str
            Sublayer and segment representing distraction trials.

        distract_list : numpy.array
            Distraction item indices relative to the segment.

        distract_B : float or numpy.array
            Context updating rate for each distraction event before
            each study event. If an array, must be of length
            n_items + 1. Distraction will not be presented on trials i
            where distract_B[i] is zero.
        
        retention_B : float
            Context updating rate for distraction after the last study
            event.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        param = prepare_study_param(
            item_list.shape[0], len(sublayers), B, Lfc, Lcf, distract_B, retention_B
        )

        # get unit indices
        f_ind = self.get_segment('f', *segment)
        item_ind = (f_ind[0] + item_list).astype(np.dtype('i'))
        f_ind_distract = self.get_segment('f', *distract_segment)
        distract_ind = (f_ind_distract[0] + distract_list).astype(np.dtype('i'))
        c_ind = self.get_sublayers('c', sublayers)

        operations.study_distract(
            self.w_fc_exp,
            self.w_fc_pre,
            self.w_cf_exp,
            self.c,
            self.c_in,
            self.f,
            item_ind,
            c_ind,
            param['B'],
            param['Lfc'],
            param['Lcf'],
            distract_ind,
            param['distract_B'],
        )

    def record_study(
        self, 
        segment, 
        item_list, 
        sublayers, 
        B, 
        Lfc, 
        Lcf, 
        distract_segment=None, 
        distract_list=None, 
        distract_B=0, 
        retention_B=0, 
        include=None, 
        exclude=None,
    ):
        """
        Study a list of items and record network states.

        Parameters
        ----------
        segment : tuple of str, str
            Sublayer and segment of items to present.

        item_list : numpy.array
            Item indices relative to the segment.

        sublayers : str or list of str
            Sublayers of context to update.

        B : float or numpy.array
            Context updating rate. If an array, specifies a rate for
            each individual study trial.

        Lfc : float or numpy.array
            Learning rate for item to context associations. If an
            array, specifies a learning rate for each individual trial.

        Lcf : float or numpy.array
            Learning rate for context to item associations.

        distract_segment : str, str
            Sublayer and segment representing distraction trials.

        distract_list : numpy.array
            Distraction item indices relative to the segment.

        distract_B : float or numpy.array
            Context updating rate for each distraction event before
            each study event. If an array, must be of length
            n_items + 1. Distraction will not be presented on trials i
            where distract_B[i] is zero.
        
        retention_B : float
            Context updating rate for distraction after the last study
            event.

        include : list of str, optional
            Network attributes to include in the recorded states.
            Default is to include all attributes.

        exclude : list of str, optional
            Network attributes to exclude from the recorded states.

        Returns
        -------
        state : list of cymr.network.Network
            Copy of the network state after presentation of each item.
        """
        if distract_segment is not None and distract_list is not None:
            distraction = True
        else:
            distraction = False
        
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        param = prepare_study_param(
            item_list.shape[0], len(sublayers), B, Lfc, Lcf, distract_B, retention_B
        )
        state = []
        for i in range(len(item_list)):
            if distraction:
                item = (*distract_segment, distract_list[i])
                self.integrate(item, sublayers, param['distract_B'][i])

            item = (*segment, item_list[i])
            self.present(
                item, sublayers, param['B'][i], param['Lfc'][i], param['Lcf'][i]
            )
            state.append(self.copy(include=include, exclude=exclude))
        
        if distraction:
            item = (*distract_segment, distract_list[-1])
            self.integrate(item, sublayers, param['distract_B'][-1])
        return state

    def record_recall(
        self,
        segment,
        recalls,
        sublayers,
        B,
        T,
        amin=0.000001,
        include=None,
        exclude=None,
    ):
        """
        Simulate a recall sequence and record network states.

        Parameters
        ----------
        segment : tuple of str, str
            Sublayer and segment from which items are recalled.

        recalls : numpy.array
            Index of each recalled item relative to the segment.

        sublayers : str or list of str
            Sublayer(s) of context to update.

        B : float or numpy.array
            Context updating rate. If an array, specifies a rate for
            each individual recall trial.

        T : float
            Decision parameter for choice rule.

        amin : float, optional
            Minimum activation for each not-yet-recalled item on each
            recall attempt.

        include : list of str, optional
            Network attributes to include in the recorded states.
            Default is to include all attributes.

        exclude : list of str, optional
            Network attributes to exclude from the recorded states.

        Returns
        -------
        state : list of cymr.network.Network
            Copy of the network state after each recall attempt.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        recalls = np.array(recalls, dtype=np.dtype('i'))
        n_item = recalls.shape[0]
        n_sub = len(sublayers)
        param = prepare_recall_param(n_item, n_sub, B, T, amin)

        f_ind = self.get_segment('f', *segment)
        exclude_items = np.zeros(self.n_f, dtype=np.dtype('i'))
        state = []
        for i, recall in enumerate(recalls):
            # calculate item support
            operations.cue_item(
                f_ind[0],
                n_item,
                self.w_cf_pre,
                self.w_cf_exp,
                self.w_ff_pre,
                self.w_ff_exp,
                self.f_in,
                self.c,
                exclude_items,
                recalls,
                i,
            )
            operations.apply_softmax(
                f_ind[0], n_item, self.f_in, exclude_items, amin, param['T']
            )

            # update the item layer
            ind = f_ind[0] + recall
            self.f[:] = 0
            self.f[ind] = 1
            exclude_items[ind] = 1

            # save context cue and the item it cued
            state.append(self.copy(include=include, exclude=exclude))

            # update context
            item = (*segment, recall)
            self.integrate(item, sublayers, param['B'][i])
        return state

    def p_recall(self, segment, recalls, sublayers, B, T, p_stop, amin=0.000001):
        """
        Calculate the probability of a specific recall sequence.

        Parameters
        ----------
        segment : tuple of str, str
            Sublayer and segment from which items are recalled.

        recalls : numpy.array
            Index of each recalled item relative to the segment.

        sublayers : str or list of str
            Sublayer(s) of context to update.

        B : float
            Context updating rate after each recalled item.

        T : float
            Decision parameter for choice rule.

        p_stop : numpy.array
            Probability of stopping recall at each output position.

        amin : float, optional
            Minimum activation for each not-yet-recalled item on each
            recall attempt.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        recalls = np.array(recalls, dtype=np.dtype('i'))
        param = prepare_recall_param(recalls.shape[0], len(sublayers), B, T, amin)

        f_ind = self.get_segment('f', *segment)
        start = f_ind[0]
        n_f = f_ind[1] - f_ind[0]
        c_ind = self.get_sublayers('c', sublayers)

        exclude = np.zeros(n_f, dtype=np.dtype('i'))
        p = np.zeros(len(recalls) + 1)
        operations.p_recall(
            start,
            n_f,
            recalls,
            self.w_fc_exp,
            self.w_fc_pre,
            self.w_cf_exp,
            self.w_cf_pre,
            self.w_ff_exp,
            self.w_ff_pre,
            self.f,
            self.f_in,
            self.c,
            self.c_in,
            c_ind,
            exclude,
            amin,
            param['B'],
            param['T'],
            p_stop,
            p,
        )
        return p

    def generate_recall(
        self,
        segment,
        sublayers,
        B,
        T,
        p_stop,
        amin=0.000001,
        filter_recalls=False,
        A1=None,
        A2=None,
    ):
        """
        Generate a sequence of simulated free recall events.

        Parameters
        ----------
        segment : tuple of str, str
            Sublayer and segment to retrieve items from.

        sublayers : str or list of str
            Sublayer(s) of context to update.

        B : float or numpy.ndarray
            Context updating rate after each recall.

        T : float
            Decision parameter for choice rule.

        p_stop : numpy.array
            Probability of stopping at each output position.

        amin : float, optional
            Minimum activation of each not-yet-recalled item on each
            recall attempt.

        filter_recalls : bool, optional
            If true, potential recalls will be filtered based on match
            to context.

        A1 : float, optional
            Intercept mapping context match to an expit to determine
            recovery probability.

        A2 : float, optional
            Slope mapping match to an expit.

        Returns
        -------
        recalls : list of int
            Indices of items recalled at each output position.
        """
        if not isinstance(sublayers, list):
            sublayers = [sublayers]

        # weights to use for recall (assume fixed during recall)
        rec_ind = self.get_segment('f', *segment)
        n_item = rec_ind[1] - rec_ind[0]
        n_sub = len(sublayers)
        param = prepare_recall_param(n_item, n_sub, B, T, amin)

        recalls = []
        exclude = np.zeros(n_item, dtype=np.dtype('i'))
        item_ind = np.arange(n_item)
        for i in range(n_item):
            # stop recall with some probability
            if np.random.rand() < p_stop[i]:
                break

            # calculate item support
            operations.cue_item(
                rec_ind[0],
                n_item,
                self.w_cf_pre,
                self.w_cf_exp,
                self.w_ff_pre,
                self.w_ff_exp,
                self.f_in,
                self.c,
                exclude,
                np.asarray(recalls, dtype=np.dtype('i')),
                i,
            )
            operations.apply_softmax(
                rec_ind[0], n_item, self.f_in, exclude, amin, param['T']
            )

            # select item for recall proportionate to support
            support = self.f_in[rec_ind[0] : rec_ind[1]]
            p_recall = support / np.sum(support)

            if filter_recalls:
                # acceptance probability based on context match
                operations.item_match(
                    rec_ind[0], n_item, self.w_fc_pre, self.w_fc_exp, self.c, self.match
                )
                operations.apply_expit(rec_ind[0], n_item, self.match, A1, A2)

                # recall probabiity is selection + acceptance
                p_recall = p_recall * self.match[rec_ind[0] : rec_ind[1]]

                # rescale probability to vary between 0 and 1
                p_recall = p_recall / np.sum(p_recall)

            if np.any(np.isnan(p_recall)):
                n = np.count_nonzero(exclude == 0)
                p_recall[exclude == 0] = 1 / n
                p_recall[exclude == 1] = 0
                recall = np.random.choice(item_ind, p=p_recall)
            else:
                recall = np.random.choice(item_ind, p=p_recall)
            recalls.append(recall)
            exclude[recall] = 1

            # integrate context associated with the item into context
            item = (*segment, recall)
            self.integrate(item, sublayers, param['B'][i])
        return recalls

    def generate_recall_lba(self, segment, sublayers, time_limit, B, A, b, s, tau):
        """Generate timed free recall using the LBA model."""
        if not isinstance(sublayers, list):
            sublayers = [sublayers]
        rec_ind = self.get_segment('f', *segment)
        n_item = rec_ind[1] - rec_ind[0]
        n_sub = len(sublayers)
        B = expand_param(B, (n_item, n_sub))

        recalls = []
        times = []
        exclude = np.zeros(n_item, dtype=np.dtype('i'))
        t = 0
        for i in range(n_item):
            if t >= time_limit:
                break

            # calculate item support
            operations.cue_item(
                rec_ind[0],
                n_item,
                self.w_cf_pre,
                self.w_cf_exp,
                self.w_ff_pre,
                self.w_ff_exp,
                self.f_in,
                self.c,
                exclude,
                np.asarray(recalls, dtype=np.dtype('i')),
                i,
            )
            support = self.f_in[rec_ind[0] : rec_ind[1]]

            # simulate response competition
            recall, irt = sample_response_lba(A, b, support, s, tau)
            t += irt

            if t <= time_limit:
                # if response happened in time, count it
                recalls.append(recall)
                times.append(t)
                exclude[recall] = 1

                # integrate context associated with the item into context
                item = (*segment, recall)
                self.integrate(item, sublayers, B[i])
        return recalls, times

    def plot(self, ax=None):
        """Plot the current state of the network."""
        if ax is None:
            _, ax = init_plot()
        for name, h in ax.items():
            if not hasattr(self, name):
                continue
            mat = getattr(self, name)
            if mat.ndim == 1:
                mat = mat[None, :]
            h.matshow(mat)
            h.set_aspect('auto')
            h.set_axis_off()
            h.set_title(name, fontsize=12)
